"""Tests for the download subcommand of nf-core tools"""

import logging
import os
import re
import shutil
import tempfile
import unittest
from pathlib import Path
from typing import List
from unittest import mock

import pytest

import nf_core.pipelines.create.create
import nf_core.pipelines.list
import nf_core.utils
from nf_core.pipelines.download import ContainerError, DownloadWorkflow, WorkflowRepo
from nf_core.synced_repo import SyncedRepo
from nf_core.utils import run_cmd

from ..utils import TEST_DATA_DIR, with_temporary_folder


class DownloadTest(unittest.TestCase):
    @pytest.fixture(autouse=True)
    def use_caplog(self, caplog):
        self._caplog = caplog

    @property
    def logged_levels(self) -> List[str]:
        return [record.levelname for record in self._caplog.records]

    @property
    def logged_messages(self) -> List[str]:
        return [record.message for record in self._caplog.records]

    def __contains__(self, item: str) -> bool:
        """Allows to check for log messages easily using the in operator inside a test:
        assert 'my log message' in self
        """
        return any(record.message == item for record in self._caplog.records if self._caplog)

    #
    # Tests for 'get_release_hash'
    #
    def test_get_release_hash_release(self):
        wfs = nf_core.pipelines.list.Workflows()
        wfs.get_remote_workflows()
        pipeline = "methylseq"
        download_obj = DownloadWorkflow(pipeline=pipeline, revision="1.6")
        (
            download_obj.pipeline,
            download_obj.wf_revisions,
            download_obj.wf_branches,
        ) = nf_core.utils.get_repo_releases_branches(pipeline, wfs)
        download_obj.get_revision_hash()
        assert download_obj.wf_sha[download_obj.revision[0]] == "b3e5e3b95aaf01d98391a62a10a3990c0a4de395"
        assert download_obj.outdir == "nf-core-methylseq_1.6"
        assert (
            download_obj.wf_download_url[download_obj.revision[0]]
            == "https://github.com/nf-core/methylseq/archive/b3e5e3b95aaf01d98391a62a10a3990c0a4de395.zip"
        )

    def test_get_release_hash_branch(self):
        wfs = nf_core.pipelines.list.Workflows()
        wfs.get_remote_workflows()
        # Exoseq pipeline is archived, so `dev` branch should be stable
        pipeline = "exoseq"
        download_obj = DownloadWorkflow(pipeline=pipeline, revision="dev")
        (
            download_obj.pipeline,
            download_obj.wf_revisions,
            download_obj.wf_branches,
        ) = nf_core.utils.get_repo_releases_branches(pipeline, wfs)
        download_obj.get_revision_hash()
        assert download_obj.wf_sha[download_obj.revision[0]] == "819cbac792b76cf66c840b567ed0ee9a2f620db7"
        assert download_obj.outdir == "nf-core-exoseq_dev"
        assert (
            download_obj.wf_download_url[download_obj.revision[0]]
            == "https://github.com/nf-core/exoseq/archive/819cbac792b76cf66c840b567ed0ee9a2f620db7.zip"
        )

    def test_get_release_hash_non_existent_release(self):
        wfs = nf_core.pipelines.list.Workflows()
        wfs.get_remote_workflows()
        pipeline = "methylseq"
        download_obj = DownloadWorkflow(pipeline=pipeline, revision="thisisfake")
        (
            download_obj.pipeline,
            download_obj.wf_revisions,
            download_obj.wf_branches,
        ) = nf_core.utils.get_repo_releases_branches(pipeline, wfs)
        with pytest.raises(AssertionError):
            download_obj.get_revision_hash()

    #
    # Tests for 'download_wf_files'
    #
    @with_temporary_folder
    def test_download_wf_files(self, outdir):
        download_obj = DownloadWorkflow(pipeline="nf-core/methylseq", revision="1.6")
        download_obj.outdir = outdir
        download_obj.wf_sha = {"1.6": "b3e5e3b95aaf01d98391a62a10a3990c0a4de395"}
        download_obj.wf_download_url = {
            "1.6": "https://github.com/nf-core/methylseq/archive/b3e5e3b95aaf01d98391a62a10a3990c0a4de395.zip"
        }
        rev = download_obj.download_wf_files(
            download_obj.revision[0],
            download_obj.wf_sha[download_obj.revision[0]],
            download_obj.wf_download_url[download_obj.revision[0]],
        )
        assert os.path.exists(os.path.join(outdir, rev, "main.nf"))

    #
    # Tests for 'download_configs'
    #
    @with_temporary_folder
    def test_download_configs(self, outdir):
        download_obj = DownloadWorkflow(pipeline="nf-core/methylseq", revision="1.6")
        download_obj.outdir = outdir
        download_obj.download_configs()
        assert os.path.exists(os.path.join(outdir, "configs", "nfcore_custom.config"))

    #
    # Tests for 'wf_use_local_configs'
    #
    @with_temporary_folder
    def test_wf_use_local_configs(self, tmp_path):
        # Get a workflow and configs
        test_pipeline_dir = os.path.join(tmp_path, "nf-core-testpipeline")
        create_obj = nf_core.pipelines.create.create.PipelineCreate(
            "testpipeline",
            "This is a test pipeline",
            "Test McTestFace",
            no_git=True,
            outdir=test_pipeline_dir,
        )
        create_obj.init_pipeline()

        with tempfile.TemporaryDirectory() as test_outdir:
            download_obj = DownloadWorkflow(pipeline="dummy", revision="1.2.0", outdir=test_outdir)
            shutil.copytree(test_pipeline_dir, Path(test_outdir, "workflow"))
            download_obj.download_configs()

            # Test the function
            download_obj.wf_use_local_configs("workflow")
            wf_config = nf_core.utils.fetch_wf_config(Path(test_outdir, "workflow"), cache_config=False)
            assert wf_config["params.custom_config_base"] == f"{test_outdir}/workflow/../configs/"

    #
    # Tests for 'find_container_images'
    #
    @with_temporary_folder
    @mock.patch("nf_core.utils.fetch_wf_config")
    def test_find_container_images_config_basic(self, tmp_path, mock_fetch_wf_config):
        download_obj = DownloadWorkflow(pipeline="dummy", outdir=tmp_path)
        mock_fetch_wf_config.return_value = {
            "process.mapping.container": "cutting-edge-container",
            "process.nocontainer": "not-so-cutting-edge",
        }
        download_obj.find_container_images("workflow")
        assert len(download_obj.containers) == 1
        assert download_obj.containers[0] == "cutting-edge-container"

    #
    # Test for 'find_container_images' in config with nextflow
    #
    @pytest.mark.skipif(
        shutil.which("nextflow") is None,
        reason="Can't run test that requires nextflow to run if not installed.",
    )
    @with_temporary_folder
    @mock.patch("nf_core.utils.fetch_wf_config")
    def test__find_container_images_config_nextflow(self, tmp_path, mock_fetch_wf_config):
        download_obj = DownloadWorkflow(pipeline="dummy", outdir=tmp_path)
        result = run_cmd("nextflow", f"config -flat {TEST_DATA_DIR}'/mock_config_containers'")
        if result is not None:
            nfconfig_raw, _ = result
            config = {}
            for line in nfconfig_raw.splitlines():
                ul = line.decode("utf-8")
                try:
                    k, v = ul.split(" = ", 1)
                    config[k] = v.strip("'\"")
                except ValueError:
                    pass
            mock_fetch_wf_config.return_value = config
            download_obj.find_container_images("workflow")
            assert len(download_obj.containers) == 4
            assert "nfcore/methylseq:1.0" in download_obj.containers
            assert "nfcore/methylseq:1.4" in download_obj.containers
            assert "nfcore/sarek:dev" in download_obj.containers
            assert (
                "https://depot.galaxyproject.org/singularity/r-shinyngs:1.7.1--r42hdfd78af_1" in download_obj.containers
            )
            # does not yet pick up nfcore/sarekvep:dev.${params.genome}, because that is no valid URL or Docker URI.

    #
    # Test for 'find_container_images' in modules
    #
    @with_temporary_folder
    @mock.patch("nf_core.utils.fetch_wf_config")
    def test_find_container_images_modules(self, tmp_path, mock_fetch_wf_config):
        download_obj = DownloadWorkflow(pipeline="dummy", outdir=tmp_path)
        mock_fetch_wf_config.return_value = {}
        download_obj.find_container_images(str(Path(TEST_DATA_DIR, "mock_module_containers")))

        # mock_docker_single_quay_io.nf
        assert "quay.io/biocontainers/singlequay:1.9--pyh9f0ad1d_0" in download_obj.containers

        # mock_dsl2_apptainer_var1.nf (possible future convention?)
        assert (
            "https://depot.galaxyproject.org/singularity/dsltwoapptainervarone:1.1.0--py38h7be5676_2"
            in download_obj.containers
        )
        assert "biocontainers/dsltwoapptainervarone:1.1.0--py38h7be5676_2" not in download_obj.containers

        # mock_dsl2_apptainer_var2.nf (possible future convention?)
        assert (
            "https://depot.galaxyproject.org/singularity/dsltwoapptainervartwo:1.1.0--hdfd78af_0"
            in download_obj.containers
        )
        assert "biocontainers/dsltwoapptainervartwo:1.1.0--hdfd78af_0" not in download_obj.containers

        # mock_dsl2_current_inverted.nf (new implementation supports if the direct download URL is listed after Docker URI)
        assert (
            "https://depot.galaxyproject.org/singularity/dsltwocurrentinv:3.3.2--h1b792b2_1" in download_obj.containers
        )
        assert "biocontainers/dsltwocurrentinv:3.3.2--h1b792b2_1" not in download_obj.containers

        # mock_dsl2_current.nf (main nf-core convention, should be the one in far the most modules)
        assert (
            "https://depot.galaxyproject.org/singularity/dsltwocurrent:1.2.1--pyhdfd78af_0" in download_obj.containers
        )
        assert "biocontainers/dsltwocurrent:1.2.1--pyhdfd78af_0" not in download_obj.containers

        # mock_dsl2_old.nf (initial DSL2 convention)
        assert "https://depot.galaxyproject.org/singularity/dsltwoold:0.23.0--0" in download_obj.containers
        assert "quay.io/biocontainers/dsltwoold:0.23.0--0" not in download_obj.containers

        # mock_dsl2_variable.nf (currently the edgiest edge case supported)
        assert (
            "https://depot.galaxyproject.org/singularity/mulled-v2-1fa26d1ce03c295fe2fdcf85831a92fbcbd7e8c2:59cdd445419f14abac76b31dd0d71217994cbcc9-0"
            in download_obj.containers
        )
        assert (
            "https://depot.galaxyproject.org/singularity/mulled-v2-1fa26d1ce03c295fe2fdcf85831a92fbcbd7e8c2:afaaa4c6f5b308b4b6aa2dd8e99e1466b2a6b0cd-0"
            in download_obj.containers
        )
        assert (
            "quay.io/biocontainers/mulled-v2-1fa26d1ce03c295fe2fdcf85831a92fbcbd7e8c2:59cdd445419f14abac76b31dd0d71217994cbcc9-0"
            not in download_obj.containers
        )
        assert (
            "quay.io/biocontainers/mulled-v2-1fa26d1ce03c295fe2fdcf85831a92fbcbd7e8c2:afaaa4c6f5b308b4b6aa2dd8e99e1466b2a6b0cd-0"
            not in download_obj.containers
        )

        # mock_seqera_container_oras.nf
        assert "oras://community.wave.seqera.io/library/umi-transfer:1.0.0--e5b0c1a65b8173b6" in download_obj.containers
        assert "community.wave.seqera.io/library/umi-transfer:1.0.0--d30e8812ea280fa1" not in download_obj.containers

        # mock_seqera_container_oras_mulled.nf
        assert (
            "oras://community.wave.seqera.io/library/umi-transfer_umicollapse:796a995ff53da9e3"
            in download_obj.containers
        )
        assert (
            "community.wave.seqera.io/library/umi-transfer_umicollapse:3298d4f1b49e33bd" not in download_obj.containers
        )

        # mock_seqera_container_http.nf
        assert (
            "https://community-cr-prod.seqera.io/docker/registry/v2/blobs/sha256/c2/c262fc09eca59edb5a724080eeceb00fb06396f510aefb229c2d2c6897e63975/data"
            in download_obj.containers
        )

        # ToDO: This URI should actually NOT be in there, but prioritize_direct_download() can not handle this case.
        #
        # It works purely by comparing the strings, thus can establish the equivalence of 'https://depot.galaxyproject.org/singularity/umi_tools:1.1.5--py39hf95cd2a_0'
        # and 'biocontainers/umi_tools:1.1.5--py39hf95cd2a_0' because of the identical string 'umi_tools:1.1.5--py39hf95cd2a_0', but has no means to establish that
        # 'https://community-cr-prod.seqera.io/docker/registry/v2/blobs/sha256/c2/c262fc09eca59edb5a724080eeceb00fb06396f510aefb229c2d2c6897e63975/data' and
        # 'community.wave.seqera.io/library/coreutils:9.5--ae99c88a9b28c264' are the equivalent container. It would need to query an API at Seqera for that.

        assert "community.wave.seqera.io/library/coreutils:9.5--ae99c88a9b28c264" in download_obj.containers

    #
    # Test for 'prioritize_direct_download'
    #
    @with_temporary_folder
    def test_prioritize_direct_download(self, tmp_path):
        download_obj = DownloadWorkflow(pipeline="dummy", outdir=tmp_path)

        # tests deduplication and https priority as well as Seqera Container exception

        test_container = [
            "https://depot.galaxyproject.org/singularity/ubuntu:22.04",
            "nf-core/ubuntu:22.04",
            "biocontainers/umi-transfer:1.5.0--h715e4b3_0",
            "https://depot.galaxyproject.org/singularity/umi-transfer:1.5.0--h715e4b3_0",
            "biocontainers/umi-transfer:1.5.0--h715e4b3_0",
            "quay.io/nf-core/sortmerna:4.3.7--6502243397c065ba",
            "nf-core/sortmerna:4.3.7--6502243397c065ba",
            "https://depot.galaxyproject.org/singularity/sortmerna:4.3.7--hdbdd923_1",
            "https://depot.galaxyproject.org/singularity/sortmerna:4.3.7--hdbdd923_0",
            "https://depot.galaxyproject.org/singularity/sortmerna:4.2.0--h9ee0642_1",
            "https://community-cr-prod.seqera.io/docker/registry/v2/blobs/sha256/63/6397750e9730a3fbcc5b4c43f14bd141c64c723fd7dad80e47921a68a7c3cd21/data",
            "https://community-cr-prod.seqera.io/docker/registry/v2/blobs/sha256/c2/c262fc09eca59edb5a724080eeceb00fb06396f510aefb229c2d2c6897e63975/data",
            "https://community-cr-prod.seqera.io/docker/registry/v2/blobs/sha256/c2/c262fc09eca59edb5a724080eeceb00fb06396f510aefb229c2d2c6897e63975/data",
        ]

        result = download_obj.prioritize_direct_download(test_container)

        # Verify that the priority works for regular https downloads (https encountered first)
        assert "https://depot.galaxyproject.org/singularity/ubuntu:22.04" in result
        assert "nf-core/ubuntu:22.04" not in result

        # Verify that the priority works for regular https downloads (https encountered second)
        assert "biocontainers/umi-transfer:1.5.0--h715e4b3_0" not in result
        assert "https://depot.galaxyproject.org/singularity/umi-transfer:1.5.0--h715e4b3_0" in result

        # Verify that the priority works for images with and without explicit registry
        # No priority here, though - the first is retained.
        assert "nf-core/sortmerna:4.3.7--6502243397c065ba" in result
        assert "quay.io/nf-core/sortmerna:4.3.7--6502243397c065ba" not in result

        # Verify that different versions of the same tool and different build numbers are retained
        assert "https://depot.galaxyproject.org/singularity/sortmerna:4.3.7--hdbdd923_1" in result
        assert "https://depot.galaxyproject.org/singularity/sortmerna:4.3.7--hdbdd923_0" in result
        assert "https://depot.galaxyproject.org/singularity/sortmerna:4.2.0--h9ee0642_1" in result

        # Verify that Seqera containers are not deduplicated...
        assert (
            "https://community-cr-prod.seqera.io/docker/registry/v2/blobs/sha256/63/6397750e9730a3fbcc5b4c43f14bd141c64c723fd7dad80e47921a68a7c3cd21/data"
            in result
        )
        assert (
            "https://community-cr-prod.seqera.io/docker/registry/v2/blobs/sha256/c2/c262fc09eca59edb5a724080eeceb00fb06396f510aefb229c2d2c6897e63975/data"
            in result
        )
        # ...but identical ones are.
        assert (
            result.count(
                "https://community-cr-prod.seqera.io/docker/registry/v2/blobs/sha256/c2/c262fc09eca59edb5a724080eeceb00fb06396f510aefb229c2d2c6897e63975/data"
            )
            == 1
        )

    #
    # Test for 'reconcile_seqera_container_uris'
    #
    @with_temporary_folder
    def test_reconcile_seqera_container_uris(self, tmp_path):
        download_obj = DownloadWorkflow(pipeline="dummy", outdir=tmp_path)

        prioritized_container = [
            "oras://community.wave.seqera.io/library/umi-transfer:1.0.0--e5b0c1a65b8173b6",
            "oras://community.wave.seqera.io/library/sylph:0.6.1--b97274cdc1caa649",
        ]

        test_container = [
            "https://depot.galaxyproject.org/singularity/ubuntu:22.04",
            "nf-core/ubuntu:22.04",
            "nf-core/ubuntu:22.04",
            "nf-core/ubuntu:22.04",
            "community.wave.seqera.io/library/umi-transfer:1.5.0--73c1a6b65e5b0b81",
            "community.wave.seqera.io/library/sylph:0.6.1--a21713a57a65a373",
            "biocontainers/sylph:0.6.1--b97274cdc1caa649",
        ]

        # test that the test_container list is returned as it is, if no prioritized_containers are specified
        result_empty = download_obj.reconcile_seqera_container_uris([], test_container)
        assert result_empty == test_container

        result = download_obj.reconcile_seqera_container_uris(prioritized_container, test_container)

        # Verify that unrelated images are retained
        assert "https://depot.galaxyproject.org/singularity/ubuntu:22.04" in result
        assert "nf-core/ubuntu:22.04" in result

        # Verify that the priority works for regular Seqera container (Native Singularity over Docker, but only for Seqera registry)
        assert "oras://community.wave.seqera.io/library/sylph:0.6.1--b97274cdc1caa649" in result
        assert "community.wave.seqera.io/library/sylph:0.6.1--a21713a57a65a373" not in result
        assert "biocontainers/sylph:0.6.1--b97274cdc1caa649" in result

        # Verify that version strings are respected: Version 1.0.0 does not replace version 1.5.0
        assert "oras://community.wave.seqera.io/library/umi-transfer:1.0.0--e5b0c1a65b8173b6" in result
        assert "community.wave.seqera.io/library/umi-transfer:1.5.0--73c1a6b65e5b0b81" in result

        # assert that the deduplication works
        assert test_container.count("nf-core/ubuntu:22.04") == 3
        assert result.count("nf-core/ubuntu:22.04") == 1

    #
    # Tests for 'singularity_pull_image'
    #
    # If Singularity is installed, but the container can't be accessed because it does not exist or there are access
    # restrictions, a RuntimeWarning is raised due to the unavailability of the image.
    @pytest.mark.skipif(
        shutil.which("singularity") is None,
        reason="Can't test what Singularity does if it's not installed.",
    )
    @with_temporary_folder
    @mock.patch("rich.progress.Progress.add_task")
    def test_singularity_pull_image_singularity_installed(self, tmp_dir, mock_rich_progress):
        download_obj = DownloadWorkflow(pipeline="dummy", outdir=tmp_dir)

        # Test successful pull
        download_obj.singularity_pull_image(
            "hello-world", f"{tmp_dir}/hello-world.sif", None, "docker.io", mock_rich_progress
        )

        # Pull again, but now the image already exists
        with pytest.raises(ContainerError.ImageExistsError):
            download_obj.singularity_pull_image(
                "hello-world", f"{tmp_dir}/hello-world.sif", None, "docker.io", mock_rich_progress
            )

        # Test successful pull with absolute URI (use tiny 3.5MB test container from the "Kogia" project: https://github.com/bschiffthaler/kogia)
        download_obj.singularity_pull_image(
            "docker.io/bschiffthaler/sed", f"{tmp_dir}/sed.sif", None, "docker.io", mock_rich_progress
        )

        # Test successful pull with absolute oras:// URI
        download_obj.singularity_pull_image(
            "oras://community.wave.seqera.io/library/umi-transfer:1.0.0--e5b0c1a65b8173b6",
            f"{tmp_dir}/umi-transfer-oras.sif",
            None,
            "docker.io",
            mock_rich_progress,
        )

        # try pulling Docker container image with oras://
        with pytest.raises(ContainerError.NoSingularityContainerError):
            download_obj.singularity_pull_image(
                "oras://ghcr.io/matthiaszepper/umi-transfer:dev",
                f"{tmp_dir}/umi-transfer-oras_impostor.sif",
                None,
                "docker.io",
                mock_rich_progress,
            )

        # try to pull from non-existing registry (Name change hello-world_new.sif is needed, otherwise ImageExistsError is raised before attempting to pull.)
        with pytest.raises(ContainerError.RegistryNotFoundError):
            download_obj.singularity_pull_image(
                "hello-world",
                f"{tmp_dir}/break_the_registry_test.sif",
                None,
                "register-this-domain-to-break-the-test.io",
                mock_rich_progress,
            )

        # test Image not found for several registries
        with pytest.raises(ContainerError.ImageNotFoundError):
            download_obj.singularity_pull_image(
                "a-container", f"{tmp_dir}/acontainer.sif", None, "quay.io", mock_rich_progress
            )

        with pytest.raises(ContainerError.ImageNotFoundError):
            download_obj.singularity_pull_image(
                "a-container", f"{tmp_dir}/acontainer.sif", None, "docker.io", mock_rich_progress
            )

        with pytest.raises(ContainerError.ImageNotFoundError):
            download_obj.singularity_pull_image(
                "a-container", f"{tmp_dir}/acontainer.sif", None, "ghcr.io", mock_rich_progress
            )

        # test Image not found for absolute URI.
        with pytest.raises(ContainerError.ImageNotFoundError):
            download_obj.singularity_pull_image(
                "docker.io/bschiffthaler/nothingtopullhere",
                f"{tmp_dir}/nothingtopullhere.sif",
                None,
                "docker.io",
                mock_rich_progress,
            )

        # Traffic from Github Actions to GitHub's Container Registry is unlimited, so no harm should be done here.
        with pytest.raises(ContainerError.InvalidTagError):
            download_obj.singularity_pull_image(
                "ewels/multiqc:go-rewrite",
                f"{tmp_dir}/multiqc-go.sif",
                None,
                "ghcr.io",
                mock_rich_progress,
            )

    @pytest.mark.skipif(
        shutil.which("singularity") is None,
        reason="Can't test what Singularity does if it's not installed.",
    )
    @with_temporary_folder
    @mock.patch("rich.progress.Progress.add_task")
    def test_singularity_pull_image_successfully(self, tmp_dir, mock_rich_progress):
        download_obj = DownloadWorkflow(pipeline="dummy", outdir=tmp_dir)
        download_obj.singularity_pull_image(
            "hello-world", f"{tmp_dir}/yet-another-hello-world.sif", None, "docker.io", mock_rich_progress
        )

    #
    # Tests for 'get_singularity_images'
    #
    @pytest.mark.skipif(
        shutil.which("singularity") is None,
        reason="Can't test what Singularity does if it's not installed.",
    )
    @with_temporary_folder
    @mock.patch("nf_core.utils.fetch_wf_config")
    def test_get_singularity_images(self, tmp_path, mock_fetch_wf_config):
        download_obj = DownloadWorkflow(
            pipeline="dummy",
            outdir=tmp_path,
            container_library=("mirage-the-imaginative-registry.io", "quay.io", "ghcr.io", "docker.io"),
        )
        mock_fetch_wf_config.return_value = {
            "process.helloworld.container": "helloworld",
            "process.hellooworld.container": "helloooooooworld",
            "process.mapping.container": "ewels/multiqc:gorewrite",
        }
        download_obj.find_container_images("workflow")
        assert len(download_obj.container_library) == 4
        # This list of fake container images should produce all kinds of ContainerErrors.
        # Test that they are all caught inside get_singularity_images().
        download_obj.get_singularity_images()

    @with_temporary_folder
    @mock.patch("os.makedirs")
    @mock.patch("os.symlink")
    @mock.patch("os.open")
    @mock.patch("os.close")
    @mock.patch("os.path.basename")
    @mock.patch("os.path.dirname")
    def test_symlink_singularity_images(
        self,
        tmp_path,
        mock_dirname,
        mock_basename,
        mock_close,
        mock_open,
        mock_symlink,
        mock_makedirs,
    ):
        # Setup
        mock_dirname.return_value = f"{tmp_path}/path/to"
        mock_basename.return_value = "singularity-image.img"
        mock_open.return_value = 12  # file descriptor
        mock_close.return_value = 12  # file descriptor

        download_obj = DownloadWorkflow(
            pipeline="dummy",
            outdir=tmp_path,
            container_library=(
                "quay.io",
                "community-cr-prod.seqera.io/docker/registry/v2",
                "depot.galaxyproject.org/singularity",
            ),
        )

        # Call the method
        download_obj.symlink_singularity_images(f"{tmp_path}/path/to/singularity-image.img")

        # Check that os.makedirs was called with the correct arguments
        mock_makedirs.assert_any_call(f"{tmp_path}/path/to", exist_ok=True)

        # Check that os.open was called with the correct arguments
        mock_open.assert_any_call(f"{tmp_path}/path/to", os.O_RDONLY)

        # Check that os.symlink was called with the correct arguments
        expected_calls = [
            mock.call(
                "./singularity-image.img",
                "./quay.io-singularity-image.img",
                dir_fd=12,
            ),
            mock.call(
                "./singularity-image.img",
                "./community-cr-prod.seqera.io-docker-registry-v2-singularity-image.img",
                dir_fd=12,
            ),
            mock.call(
                "./singularity-image.img",
                "./depot.galaxyproject.org-singularity-singularity-image.img",
                dir_fd=12,
            ),
        ]
        mock_symlink.assert_has_calls(expected_calls, any_order=True)

    @with_temporary_folder
    @mock.patch("os.makedirs")
    @mock.patch("os.symlink")
    @mock.patch("os.open")
    @mock.patch("os.close")
    @mock.patch("re.sub")
    @mock.patch("os.path.basename")
    @mock.patch("os.path.dirname")
    def test_symlink_singularity_images_registry(
        self,
        tmp_path,
        mock_dirname,
        mock_basename,
        mock_resub,
        mock_close,
        mock_open,
        mock_symlink,
        mock_makedirs,
    ):
        # Setup
        mock_resub.return_value = "singularity-image.img"
        mock_dirname.return_value = f"{tmp_path}/path/to"
        mock_basename.return_value = "quay.io-singularity-image.img"
        mock_open.return_value = 12  # file descriptor
        mock_close.return_value = 12  # file descriptor

        download_obj = DownloadWorkflow(
            pipeline="dummy",
            outdir=tmp_path,
            container_library=("quay.io", "community-cr-prod.seqera.io/docker/registry/v2"),
        )

        download_obj.registry_set = {"quay.io", "community-cr-prod.seqera.io/docker/registry/v2"}

        # Call the method with registry - should not happen, but preserve it then.
        download_obj.symlink_singularity_images(f"{tmp_path}/path/to/quay.io-singularity-image.img")
        print(mock_resub.call_args)

        # Check that os.makedirs was called with the correct arguments
        mock_makedirs.assert_any_call(f"{tmp_path}/path/to", exist_ok=True)

        # Check that os.symlink was called with the correct arguments
        mock_symlink.assert_called_with(
            "./quay.io-singularity-image.img",
            "./community-cr-prod.seqera.io-docker-registry-v2-singularity-image.img",
            dir_fd=12,
        )
        # Check that there is no attempt to symlink to itself (test parameters would result in that behavior if not checked in the function)
        assert (
            unittest.mock.call("./quay.io-singularity-image.img", "./quay.io-singularity-image.img", dir_fd=12)
            not in mock_symlink.call_args_list
        )

        # Normally it would be called for each registry, but since quay.io is part of the name, it
        # will only be called once, as no symlink to itself must be created.
        mock_open.assert_called_once_with(f"{tmp_path}/path/to", os.O_RDONLY)

    #
    # Test for gather_registries'
    #
    @with_temporary_folder
    @mock.patch("nf_core.utils.fetch_wf_config")
    def test_gather_registries(self, tmp_path, mock_fetch_wf_config):
        download_obj = DownloadWorkflow(
            pipeline="dummy",
            outdir=tmp_path,
            container_library=None,
        )
        mock_fetch_wf_config.return_value = {
            "apptainer.registry": "apptainer-registry.io",
            "docker.registry": "docker.io",
            "podman.registry": "podman-registry.io",
            "singularity.registry": "singularity-registry.io",
            "someother.registry": "fake-registry.io",
        }
        download_obj.gather_registries(tmp_path)
        assert download_obj.registry_set
        assert isinstance(download_obj.registry_set, set)
        assert len(download_obj.registry_set) == 8

        assert "quay.io" in download_obj.registry_set  # default registry, if no container library is provided.
        assert (
            "depot.galaxyproject.org/singularity" in download_obj.registry_set
        )  # default registry, often hardcoded in modules
        assert "community.wave.seqera.io/library" in download_obj.registry_set  # Seqera containers Docker
        assert (
            "community-cr-prod.seqera.io/docker/registry/v2" in download_obj.registry_set
        )  # Seqera containers Singularity https:// download
        assert "apptainer-registry.io" in download_obj.registry_set
        assert "docker.io" in download_obj.registry_set
        assert "podman-registry.io" in download_obj.registry_set
        assert "singularity-registry.io" in download_obj.registry_set
        # it should only pull the apptainer, docker, podman and singularity registry from the config, but not any registry.
        assert "fake-registry.io" not in download_obj.registry_set

    #
    # If Singularity is not installed, it raises a OSError because the singularity command can't be found.
    #
    @pytest.mark.skipif(
        shutil.which("singularity") is not None,
        reason="Can't test how the code behaves when singularity is not installed if it is.",
    )
    @with_temporary_folder
    @mock.patch("rich.progress.Progress.add_task")
    def test_singularity_pull_image_singularity_not_installed(self, tmp_dir, mock_rich_progress):
        download_obj = DownloadWorkflow(pipeline="dummy", outdir=tmp_dir)
        with pytest.raises(OSError):
            download_obj.singularity_pull_image(
                "a-container", f"{tmp_dir}/anothercontainer.sif", None, "quay.io", mock_rich_progress
            )

    #
    # Test for 'singularity_image_filenames' function
    #
    @with_temporary_folder
    def test_singularity_image_filenames(self, tmp_path):
        os.environ["NXF_SINGULARITY_CACHEDIR"] = f"{tmp_path}/cachedir"

        download_obj = DownloadWorkflow(pipeline="dummy", outdir=tmp_path)
        download_obj.outdir = tmp_path
        download_obj.container_cache_utilisation = "amend"

        download_obj.registry_set = {
            "docker.io",
            "quay.io",
            "depot.galaxyproject.org/singularity",
            "community.wave.seqera.io/library",
            "community-cr-prod.seqera.io/docker/registry/v2",
        }

        ## Test phase I: Container not yet cached, should be amended to cache
        # out_path: str, Path to cache
        # cache_path: None

        result = download_obj.singularity_image_filenames(
            "https://depot.galaxyproject.org/singularity/bbmap:38.93--he522d1c_0"
        )

        # Assert that the result is a tuple of length 2
        self.assertIsInstance(result, tuple)
        self.assertEqual(len(result), 2)

        # Assert that the types of the elements are (str, None)
        self.assertTrue(all((isinstance(element, str), element is None) for element in result))

        # assert that the correct out_path is returned that points to the cache
        assert result[0].endswith("/cachedir/bbmap-38.93--he522d1c_0.img")

        ## Test phase II: Test various container names
        # out_path: str, Path to cache
        # cache_path: None

        # Test --- mulled containers #
        result = download_obj.singularity_image_filenames(
            "quay.io/biocontainers/mulled-v2-1fa26d1ce03c295fe2fdcf85831a92fbcbd7e8c2:59cdd445419f14abac76b31dd0d71217994cbcc9-0"
        )
        assert result[0].endswith(
            "/cachedir/biocontainers-mulled-v2-1fa26d1ce03c295fe2fdcf85831a92fbcbd7e8c2-59cdd445419f14abac76b31dd0d71217994cbcc9-0.img"
        )

        # Test --- Docker containers without registry #
        result = download_obj.singularity_image_filenames("nf-core/ubuntu:20.04")
        assert result[0].endswith("/cachedir/nf-core-ubuntu-20.04.img")

        # Test --- Docker container with explicit registry -> should be trimmed #
        result = download_obj.singularity_image_filenames("docker.io/nf-core/ubuntu:20.04")
        assert result[0].endswith("/cachedir/nf-core-ubuntu-20.04.img")

        # Test --- Docker container with explicit registry not in registry set -> can't be trimmed
        result = download_obj.singularity_image_filenames("mirage-the-imaginative-registry.io/nf-core/ubuntu:20.04")
        assert result[0].endswith("/cachedir/mirage-the-imaginative-registry.io-nf-core-ubuntu-20.04.img")

        # Test --- Seqera Docker containers: Trimmed, because it is hard-coded in the registry set.
        result = download_obj.singularity_image_filenames(
            "community.wave.seqera.io/library/coreutils:9.5--ae99c88a9b28c264"
        )
        assert result[0].endswith("/cachedir/coreutils-9.5--ae99c88a9b28c264.img")

        # Test --- Seqera Singularity containers: Trimmed, because it is hard-coded in the registry set.
        result = download_obj.singularity_image_filenames(
            "https://community-cr-prod.seqera.io/docker/registry/v2/blobs/sha256/c2/c262fc09eca59edb5a724080eeceb00fb06396f510aefb229c2d2c6897e63975/data"
        )
        assert result[0].endswith(
            "cachedir/blobs-sha256-c2-c262fc09eca59edb5a724080eeceb00fb06396f510aefb229c2d2c6897e63975-data.img"
        )

        ## Test phase III: Container will be cached but also copied to out_path
        # out_path: str, Path to cache
        # cache_path: str, Path to cache
        download_obj.container_cache_utilisation = "copy"
        result = download_obj.singularity_image_filenames(
            "https://depot.galaxyproject.org/singularity/bbmap:38.93--he522d1c_0"
        )

        self.assertTrue(all(isinstance(element, str) for element in result))
        assert result[0].endswith("/singularity-images/bbmap-38.93--he522d1c_0.img")
        assert result[1].endswith("/cachedir/bbmap-38.93--he522d1c_0.img")

        ## Test phase IV: Expect an error if no NXF_SINGULARITY_CACHEDIR is defined
        os.environ["NXF_SINGULARITY_CACHEDIR"] = ""
        with self.assertRaises(FileNotFoundError):
            download_obj.singularity_image_filenames(
                "https://depot.galaxyproject.org/singularity/bbmap:38.93--he522d1c_0"
            )

    #
    # Test for '--singularity-cache remote --singularity-cache-index'. Provide a list of containers already available in a remote location.
    #
    @with_temporary_folder
    def test_remote_container_functionality(self, tmp_dir):
        os.environ["NXF_SINGULARITY_CACHEDIR"] = "foo"

        download_obj = DownloadWorkflow(
            pipeline="nf-core/rnaseq",
            outdir=os.path.join(tmp_dir, "new"),
            revision="3.9",
            compress_type="none",
            container_cache_index=str(Path(TEST_DATA_DIR, "testdata_remote_containers.txt")),
        )

        download_obj.include_configs = False  # suppress prompt, because stderr.is_interactive doesn't.

        # test if the settings are changed to mandatory defaults, if an external cache index is used.
        assert download_obj.container_cache_utilisation == "remote" and download_obj.container_system == "singularity"
        assert isinstance(download_obj.containers_remote, list) and len(download_obj.containers_remote) == 0
        # read in the file
        download_obj.read_remote_containers()
        assert len(download_obj.containers_remote) == 33
        assert "depot.galaxyproject.org-singularity-salmon-1.5.2--h84f40af_0.img" in download_obj.containers_remote
        assert "MV Rena" not in download_obj.containers_remote  # decoy in test file

    #
    # Tests for the main entry method 'download_workflow'
    #
    @with_temporary_folder
    @mock.patch("nf_core.pipelines.download.DownloadWorkflow.singularity_pull_image")
    @mock.patch("shutil.which")
    def test_download_workflow_with_success(self, tmp_dir, mock_download_image, mock_singularity_installed):
        os.environ["NXF_SINGULARITY_CACHEDIR"] = "foo"

        download_obj = DownloadWorkflow(
            pipeline="nf-core/methylseq",
            outdir=os.path.join(tmp_dir, "new"),
            container_system="singularity",
            revision="1.6",
            compress_type="none",
            container_cache_utilisation="copy",
        )

        download_obj.include_configs = True  # suppress prompt, because stderr.is_interactive doesn't.
        download_obj.download_workflow()

    #
    # Test Download for Seqera Platform
    #
    @with_temporary_folder
    @mock.patch("nf_core.pipelines.download.DownloadWorkflow.get_singularity_images")
    def test_download_workflow_for_platform(self, tmp_dir, _):
        download_obj = DownloadWorkflow(
            pipeline="nf-core/rnaseq",
            revision=("3.7", "3.9"),
            compress_type="none",
            platform=True,
            container_system="singularity",
        )

        download_obj.include_configs = False  # suppress prompt, because stderr.is_interactive doesn't.

        assert isinstance(download_obj.revision, list) and len(download_obj.revision) == 2
        assert isinstance(download_obj.wf_sha, dict) and len(download_obj.wf_sha) == 0
        assert isinstance(download_obj.wf_download_url, dict) and len(download_obj.wf_download_url) == 0

        wfs = nf_core.pipelines.list.Workflows()
        wfs.get_remote_workflows()
        (
            download_obj.pipeline,
            download_obj.wf_revisions,
            download_obj.wf_branches,
        ) = nf_core.utils.get_repo_releases_branches(download_obj.pipeline, wfs)

        download_obj.get_revision_hash()

        # download_obj.wf_download_url is not set for Seqera Platform downloads, but the sha values are
        assert isinstance(download_obj.wf_sha, dict) and len(download_obj.wf_sha) == 2
        assert isinstance(download_obj.wf_download_url, dict) and len(download_obj.wf_download_url) == 0

        # The outdir for multiple revisions is the pipeline name and date: e.g. nf-core-rnaseq_2023-04-27_18-54
        assert bool(re.search(r"nf-core-rnaseq_\d{4}-\d{2}-\d{1,2}_\d{1,2}-\d{1,2}", download_obj.outdir, re.S))

        download_obj.output_filename = f"{download_obj.outdir}.git"
        download_obj.download_workflow_platform(location=tmp_dir)

        assert download_obj.workflow_repo
        assert isinstance(download_obj.workflow_repo, WorkflowRepo)
        assert issubclass(type(download_obj.workflow_repo), SyncedRepo)

        # corroborate that the other revisions are inaccessible to the user.
        all_tags = {tag.name for tag in download_obj.workflow_repo.tags}
        all_heads = {head.name for head in download_obj.workflow_repo.heads}

        assert set(download_obj.revision) == all_tags
        # assert that the download has a "latest" branch.
        assert "latest" in all_heads

        # download_obj.download_workflow_platform(location=tmp_dir) will run container image detection for all requested revisions
        assert isinstance(download_obj.containers, list) and len(download_obj.containers) == 33
        assert (
            "https://depot.galaxyproject.org/singularity/bbmap:38.93--he522d1c_0" in download_obj.containers
        )  # direct definition
        assert (
            "https://depot.galaxyproject.org/singularity/mulled-v2-1fa26d1ce03c295fe2fdcf85831a92fbcbd7e8c2:59cdd445419f14abac76b31dd0d71217994cbcc9-0"
            in download_obj.containers
        )  # indirect definition via $container variable.

        # clean-up
        # remove "nf-core-rnaseq*" directories
        for path in Path().cwd().glob("nf-core-rnaseq*"):
            shutil.rmtree(path)

    #
    # Brief test adding a single custom tag to Seqera Platform download
    #
    @mock.patch("nf_core.pipelines.download.DownloadWorkflow.get_singularity_images")
    @with_temporary_folder
    def test_download_workflow_for_platform_with_one_custom_tag(self, _, tmp_dir):
        download_obj = DownloadWorkflow(
            pipeline="nf-core/rnaseq",
            revision=("3.9"),
            compress_type="none",
            platform=True,
            container_system=None,
            additional_tags=("3.9=cool_revision",),
        )
        assert isinstance(download_obj.additional_tags, list) and len(download_obj.additional_tags) == 1

        # clean-up
        # remove "nf-core-rnaseq*" directories
        for path in Path().cwd().glob("nf-core-rnaseq*"):
            shutil.rmtree(path)

    #
    # Test adding custom tags to Seqera Platform download (full test)
    #
    @mock.patch("nf_core.pipelines.download.DownloadWorkflow.get_singularity_images")
    @with_temporary_folder
    def test_download_workflow_for_platform_with_custom_tags(self, _, tmp_dir):
        with self._caplog.at_level(logging.INFO):
            from git.refs.tag import TagReference

            download_obj = DownloadWorkflow(
                pipeline="nf-core/rnaseq",
                revision=("3.7", "3.9"),
                compress_type="none",
                platform=True,
                container_system=None,
                additional_tags=(
                    "3.7=a.tad.outdated",
                    "3.9=cool_revision",
                    "3.9=invalid tag",
                    "3.14.0=not_included",
                    "What is this?",
                ),
            )

            download_obj.include_configs = False  # suppress prompt, because stderr.is_interactive doesn't.

            assert isinstance(download_obj.revision, list) and len(download_obj.revision) == 2
            assert isinstance(download_obj.wf_sha, dict) and len(download_obj.wf_sha) == 0
            assert isinstance(download_obj.wf_download_url, dict) and len(download_obj.wf_download_url) == 0
            assert isinstance(download_obj.additional_tags, list) and len(download_obj.additional_tags) == 5

            wfs = nf_core.pipelines.list.Workflows()
            wfs.get_remote_workflows()
            (
                download_obj.pipeline,
                download_obj.wf_revisions,
                download_obj.wf_branches,
            ) = nf_core.utils.get_repo_releases_branches(download_obj.pipeline, wfs)

            download_obj.get_revision_hash()
            download_obj.output_filename = f"{download_obj.outdir}.git"
            download_obj.download_workflow_platform(location=tmp_dir)

            assert download_obj.workflow_repo
            assert isinstance(download_obj.workflow_repo, WorkflowRepo)
            assert issubclass(type(download_obj.workflow_repo), SyncedRepo)
            assert "Locally cached repository: nf-core/rnaseq, revisions 3.7, 3.9" in repr(download_obj.workflow_repo)

            # assert that every additional tag has been passed on to the WorkflowRepo instance
            assert download_obj.additional_tags == download_obj.workflow_repo.additional_tags

            # assert that the additional tags are all TagReference objects
            assert all(isinstance(tag, TagReference) for tag in download_obj.workflow_repo.tags)

            workflow_repo_tags = {tag.name for tag in download_obj.workflow_repo.tags}
            assert len(workflow_repo_tags) == 4
            # the invalid/malformed additional_tags should not have been added.
            assert all(tag in workflow_repo_tags for tag in {"3.7", "a.tad.outdated", "cool_revision", "3.9"})
            assert not any(tag in workflow_repo_tags for tag in {"invalid tag", "not_included", "What is this?"})

            assert all(
                log in self.logged_messages
                for log in {
                    "[red]Could not apply invalid `--tag` specification[/]: '3.9=invalid tag'",
                    "[red]Adding tag 'not_included' to '3.14.0' failed.[/]\n Mind that '3.14.0' must be a valid git reference that resolves to a commit.",
                    "[red]Could not apply invalid `--tag` specification[/]: 'What is this?'",
                }
            )

            # clean-up
            # remove "nf-core-rnaseq*" directories
            for path in Path().cwd().glob("nf-core-rnaseq*"):
                shutil.rmtree(path)



from nf_core.components.install import ComponentInstall


class SubworkflowInstall(ComponentInstall):
    def __init__(
        self,
        pipeline_dir,
        force=False,
        prompt=False,
        sha=None,
        remote_url=None,
        branch=None,
        no_pull=False,
        installed_by=None,
    ):
        super().__init__(
            pipeline_dir,
            "subworkflows",
            force=force,
            prompt=prompt,
            sha=sha,
            remote_url=remote_url,
            branch=branch,
            no_pull=no_pull,
            installed_by=installed_by,
        )



import logging
import os
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import questionary
import yaml
from rich import box
from rich.console import Group
from rich.markdown import Markdown
from rich.panel import Panel
from rich.syntax import Syntax
from rich.table import Table
from rich.text import Text

import nf_core.utils
from nf_core.components.components_command import ComponentCommand
from nf_core.components.components_utils import NF_CORE_MODULES_REMOTE
from nf_core.modules.modules_json import ModulesJson

log = logging.getLogger(__name__)


class ComponentInfo(ComponentCommand):
    """
    Class to print information of a module/subworkflow.

    Attributes
    ----------
    meta : YAML object
        stores the information from meta.yml file
    local_path : str
        path of the local modules/subworkflows
    remote_location : str
        remote repository URL
    local : bool
        indicates if the module/subworkflow is locally installed or not
    repo_type : str
        repository type. Can be either 'pipeline' or 'modules'
    modules_json : ModulesJson object
        contains 'modules.json' file information from a pipeline
    component_name : str
        name of the module/subworkflow to get information from

    Methods
    -------
    init_mod_name(component)
        Makes sure that we have a modules/subworkflows name
    get_component_info()
        Given the name of a module/subworkflow, parse meta.yml and print usage help
    get_local_yaml()
        Attempt to get the meta.yml file from a locally installed module/subworkflow
    get_remote_yaml()
        Attempt to get the meta.yml file from a remote repo
    generate_component_info_help()
        Take the parsed meta.yml and generate rich help
    """

    def __init__(
        self,
        component_type: str,
        pipeline_dir: Union[str, Path],
        component_name: str,
        remote_url: Optional[str] = None,
        branch: Optional[str] = None,
        no_pull: bool = False,
    ):
        super().__init__(component_type, pipeline_dir, remote_url, branch, no_pull)
        self.meta: Optional[Dict] = None
        self.local_path: Optional[Path] = None
        self.remote_location: Optional[str] = None
        self.local: bool = False
        self.modules_json: Optional[ModulesJson] = None

        if self.repo_type == "pipeline":
            # Check modules directory structure
            if self.component_type == "modules":
                self.check_modules_structure()
            # Check modules.json up to date
            self.modules_json = ModulesJson(self.directory)
            self.modules_json.check_up_to_date()
        else:
            self.modules_json = None
        self.component = self.init_mod_name(component_name)

    def _configure_repo_and_paths(self, nf_dir_req=False) -> None:
        """
        Override the default with nf_dir_req set to False to allow
        info to be run from anywhere and still return remote info
        """
        return super()._configure_repo_and_paths(nf_dir_req)

    def init_mod_name(self, component: Optional[str]) -> str:
        """
        Makes sure that we have a module/subworkflow name before proceeding.

        Args:
            module: str: Module name to check
        """
        if component is None:
            self.local = questionary.confirm(
                f"Is the {self.component_type[:-1]} locally installed?", style=nf_core.utils.nfcore_question_style
            ).unsafe_ask()
            if self.local:
                if self.repo_type == "modules":
                    components = self.get_components_clone_modules()
                elif self.repo_type == "pipeline":
                    assert self.modules_json is not None  # mypy
                    all_components: List[Tuple[str, str]] = self.modules_json.get_all_components(
                        self.component_type
                    ).get(self.modules_repo.remote_url, [])

                    components = [
                        component if directory == self.modules_repo.repo_path else f"{directory}/{component}"
                        for directory, component in all_components
                    ]
                    if not components:
                        raise UserWarning(
                            f"No {self.component_type[:-1]} installed from '{self.modules_repo.remote_url}'"
                        )
                else:
                    raise UserWarning("Unknown repository type")
            else:
                components = self.modules_repo.get_avail_components(self.component_type)
            components.sort()
            component = questionary.autocomplete(
                f"Please select a {self.component_type[:-1]}",
                choices=components,
                style=nf_core.utils.nfcore_question_style,
            ).unsafe_ask()
            while component not in components:
                log.info(f"'{component}' is not a valid {self.component_type[:-1]} name")
                component = questionary.autocomplete(
                    f"Please select a new {self.component_type[:-1]}",
                    choices=components,
                    style=nf_core.utils.nfcore_question_style,
                ).unsafe_ask()
        else:
            if self.repo_type == "pipeline":
                assert self.modules_json is not None  # mypy
                # check if the module is locally installed
                local_paths = self.modules_json.get_all_components(self.component_type).get(
                    self.modules_repo.remote_url
                )  # type: ignore
                if local_paths is not None:
                    for directory, comp in local_paths:
                        if comp == component:
                            component_base_path = Path(self.directory, self.component_type)
                            self.local_path = Path(component_base_path, directory, component)
                            break
                if self.local_path:
                    self.local = True

        return component

    def get_component_info(self):
        """Given the name of a module/subworkflow, parse meta.yml and print usage help."""

        # Running with a local install, try to find the local meta
        if self.local:
            self.meta = self.get_local_yaml()

        # Either failed locally or in remote mode
        if not self.meta:
            self.meta = self.get_remote_yaml()

        # Could not find the meta
        if self.meta is False:
            raise UserWarning(f"Could not find {self.component_type[:-1]} '{self.component}'")

        return self.generate_component_info_help()

    def get_local_yaml(self) -> Optional[Dict]:
        """Attempt to get the meta.yml file from a locally installed module/subworkflow.

        Returns:
            Optional[dict]: Parsed meta.yml if found, None otherwise
        """

        if self.repo_type == "pipeline":
            assert self.modules_json is not None  # mypy
            # Try to find and load the meta.yml file
            component_base_path = Path(self.directory, self.component_type)
            # Check that we have any modules/subworkflows installed from this repo
            components = self.modules_json.get_all_components(self.component_type).get(self.modules_repo.remote_url)
            if components is None:
                raise LookupError(f"No {self.component_type[:-1]} installed from {self.modules_repo.remote_url}")

            component_names = [component for _, component in components]

            if self.component in component_names:
                install_dir = [directory for directory, module in components if module == self.component][0]
                comp_dir = Path(component_base_path, install_dir, self.component)
                meta_fn = Path(comp_dir, "meta.yml")
                if meta_fn.exists():
                    log.debug(f"Found local file: {meta_fn}")
                    with open(meta_fn) as fh:
                        self.local_path = comp_dir
                        return yaml.safe_load(fh)

            log.debug(f"{self.component_type[:-1].title()} '{self.component}' meta.yml not found locally")
        else:
            component_base_path = Path(self.directory, self.component_type, self.org)
            if self.component in os.listdir(component_base_path):
                comp_dir = Path(component_base_path, self.component)
                meta_fn = Path(comp_dir, "meta.yml")
                if meta_fn.exists():
                    log.debug(f"Found local file: {meta_fn}")
                    with open(meta_fn) as fh:
                        self.local_path = comp_dir
                        return yaml.safe_load(fh)
            log.debug(f"{self.component_type[:-1].title()} '{self.component}' meta.yml not found locally")

        return {}

    def get_remote_yaml(self) -> Optional[Dict]:
        """Attempt to get the meta.yml file from a remote repo.

        Returns:
            dict or bool: Parsed meta.yml found, False otherwise
        """
        # Check if our requested module/subworkflow is there
        if self.component not in self.modules_repo.get_avail_components(self.component_type):
            return None

        file_contents = self.modules_repo.get_meta_yml(self.component_type, self.component)
        if file_contents is None:
            return None
        self.remote_location = self.modules_repo.remote_url
        return yaml.safe_load(file_contents)

    def generate_params_table(self, type) -> Table:
        "Generate a rich table for inputs and outputs"
        table = Table(expand=True, show_lines=True, box=box.MINIMAL_HEAVY_HEAD, padding=0)
        table.add_column(f":inbox_tray: {type}")
        table.add_column("Description")
        if self.component_type == "modules":
            table.add_column("Pattern", justify="right", style="green")
        elif self.component_type == "subworkflows":
            table.add_column("Structure", justify="right", style="green")
        return table

    def get_channel_structure(self, structure: dict) -> str:
        "Get the structure of a channel"
        structure_str = ""
        for key, info in structure.items():
            pattern = f" - {info['pattern']}" if info.get("pattern") else ""
            structure_str += f"{key} ({info['type']}{pattern})"
        return structure_str

    def generate_component_info_help(self):
        """Take the parsed meta.yml and generate rich help.

        Returns:
            rich renderable
        """

        renderables = []

        # Intro panel
        intro_text = Text()
        if self.local_path:
            intro_text.append(Text.from_markup(f"Location: [blue]{self.local_path}\n"))
        elif self.remote_location:
            intro_text.append(
                Text.from_markup(
                    ":globe_with_meridians: Repository: "
                    f"{ '[link={self.remote_location}]' if self.remote_location.startswith('http') else ''}"
                    f"{self.remote_location}"
                    f"{'[/link]' if self.remote_location.startswith('http') else '' }"
                    "\n"
                )
            )
        if self.meta is None:
            raise UserWarning("No meta.yml file found")
        if self.meta.get("tools"):
            tools_strings = []
            for tool in self.meta["tools"]:
                for tool_name, tool_meta in tool.items():
                    if "homepage" in tool_meta:
                        tools_strings.append(f"[link={tool_meta['homepage']}]{tool_name}[/link]")
                    else:
                        tools_strings.append(f"{tool_name}")
            intro_text.append(Text.from_markup(f":wrench: Tools: {', '.join(tools_strings)}\n", style="dim"))

        if self.meta.get("description"):
            intro_text.append(Text.from_markup(f":book: Description: {self.meta['description']}", style="dim"))

        renderables.append(
            Panel(
                intro_text,
                title=f"[bold]{self.component_type[:-1].title()}: [green]{self.component}\n",
                title_align="left",
            )
        )

        # Inputs
        if self.meta.get("input"):
            inputs_table = self.generate_params_table("Inputs")
            for i, input in enumerate(self.meta["input"]):
                inputs_table.add_row(f"[italic]input[{i}][/]", "", "")
                if self.component_type == "modules":
                    for element in input:
                        for key, info in element.items():
                            inputs_table.add_row(
                                f"[orange1 on black] {key} [/][dim i] ({info['type']})",
                                Markdown(info["description"] if info["description"] else ""),
                                info.get("pattern", ""),
                            )
                elif self.component_type == "subworkflows":
                    for key, info in input.items():
                        inputs_table.add_row(
                            f"[orange1 on black] {key} [/][dim i]",
                            Markdown(info["description"] if info["description"] else ""),
                            self.get_channel_structure(info["structure"]) if info.get("structure") else "",
                        )

            renderables.append(inputs_table)

        # Outputs
        if self.meta.get("output"):
            outputs_table = self.generate_params_table("Outputs")
            for output in self.meta["output"]:
                if self.component_type == "modules":
                    for ch_name, elements in output.items():
                        outputs_table.add_row(f"{ch_name}", "", "")
                        for element in elements:
                            for key, info in element.items():
                                outputs_table.add_row(
                                    f"[orange1 on black] {key} [/][dim i] ({info['type']})",
                                    Markdown(info["description"] if info["description"] else ""),
                                    info.get("pattern", ""),
                                )
                elif self.component_type == "subworkflows":
                    for key, info in output.items():
                        outputs_table.add_row(
                            f"[orange1 on black] {key} [/][dim i]",
                            Markdown(info["description"] if info["description"] else ""),
                            self.get_channel_structure(info["structure"]) if info.get("structure") else "",
                        )

            renderables.append(outputs_table)

        # Installation command
        if self.remote_location and not self.local:
            cmd_base = f"nf-core {self.component_type}"
            if self.remote_location != NF_CORE_MODULES_REMOTE:
                cmd_base = f"nf-core {self.component_type} --git-remote {self.remote_location}"
            renderables.append(
                Text.from_markup(f"\n :computer:  Installation command: [magenta]{cmd_base} install {self.component}\n")
            )

        # Print include statement
        if self.local_path and self.modules_repo.repo_path is not None:
            install_folder = Path(self.directory, self.component_type, self.modules_repo.repo_path)
            component_name = "_".join(self.component.upper().split("/"))
            renderables.append(
                Text.from_markup(f"\n [blue]Use the following statement to include this {self.component_type[:-1]}:")
            )
            renderables.append(
                Syntax(
                    f"include {{ {component_name} }} from '../{Path(install_folder, self.component).relative_to(self.directory)}/main'",
                    "groovy",
                    theme="ansi_dark",
                    padding=1,
                )
            )
            if self.component_type == "subworkflows":
                subworkflow_config = Path(install_folder, self.component, "nextflow.config").relative_to(self.directory)
                if os.path.isfile(subworkflow_config):
                    renderables.append(
                        Text.from_markup("\n [blue]Add the following config statement to use this subworkflow:")
                    )
                    renderables.append(
                        Syntax(f"includeConfig '{subworkflow_config}'", "groovy", theme="ansi_dark", padding=1)
                    )

        return Group(*renderables)



import logging
from pathlib import Path
from typing import Optional, Union

from nf_core.components.list import ComponentList

log = logging.getLogger(__name__)


class ModuleList(ComponentList):
    def __init__(
        self,
        pipeline_dir: Union[str, Path] = ".",
        remote: bool = True,
        remote_url: Optional[str] = None,
        branch: Optional[str] = None,
        no_pull: bool = False,
    ):
        super().__init__("modules", pipeline_dir, remote, remote_url, branch, no_pull)



from textual.app import ComposeResult
from textual.containers import Center
from textual.screen import Screen
from textual.widgets import Button, Footer, Header, Markdown, Static

from nf_core.utils import nfcore_logo

markdown = """
# Welcome to the nf-core pipeline creation wizard

This app will help you create a new Nextflow pipeline
from the [nf-core/tools pipeline template](https://github.com/nf-core/tools).

The template helps anyone benefit from nf-core best practices,
and is a requirement for nf-core pipelines.

> 💡 If you want to add a pipeline to nf-core, please
> [join on Slack](https://nf-co.re/join) and discuss your plans with the
> community as early as possible; _**ideally before you start on your pipeline!**_
> See the [nf-core guidelines](https://nf-co.re/docs/contributing/guidelines)
> and the [#new-pipelines](https://nfcore.slack.com/channels/new-pipelines)
> Slack channel for more information.
"""


class WelcomeScreen(Screen):
    """A welcome screen for the app."""

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        yield Static(
            "\n" + "\n".join(nfcore_logo) + "\n",
            id="logo",
        )
        yield Markdown(markdown)
        yield Center(Button("Let's go!", id="start", variant="success"), classes="cta")



import logging
import shutil
import tempfile
from pathlib import Path
from unittest import mock

import questionary
import yaml

import nf_core.utils
from nf_core.components.components_utils import NF_CORE_MODULES_NAME, NF_CORE_MODULES_REMOTE
from nf_core.modules.modules_json import ModulesJson
from nf_core.modules.update import ModuleUpdate
from nf_core.subworkflows.update import SubworkflowUpdate

from ..test_subworkflows import TestSubworkflows
from ..utils import OLD_SUBWORKFLOWS_SHA, cmp_component


class TestSubworkflowsUpdate(TestSubworkflows):
    def test_install_and_update(self):
        """Installs a subworkflow in the pipeline and updates it (no change)"""
        self.subworkflow_install.install("bam_stats_samtools")
        update_obj = SubworkflowUpdate(self.pipeline_dir, show_diff=False)

        # Copy the sw files and check that they are unaffected by the update
        tmpdir = Path(tempfile.TemporaryDirectory().name)
        sw_path = Path(self.pipeline_dir, "subworkflows", NF_CORE_MODULES_NAME, "bam_stats_samtools")
        shutil.copytree(sw_path, tmpdir)

        assert update_obj.update("bam_stats_samtools") is True
        assert cmp_component(tmpdir, sw_path) is True

    def test_install_at_hash_and_update(self):
        """Installs an old version of a subworkflow in the pipeline and updates it"""
        assert self.subworkflow_install_old.install("fastq_align_bowtie2")
        update_obj = SubworkflowUpdate(self.pipeline_dir, show_diff=False, update_deps=True)
        old_mod_json = ModulesJson(self.pipeline_dir).get_modules_json()

        # Copy the sw files and check that they are affected by the update
        tmpdir = Path(tempfile.TemporaryDirectory().name)

        sw_path = Path(self.pipeline_dir, "subworkflows", NF_CORE_MODULES_NAME, "fastq_align_bowtie2")
        shutil.copytree(sw_path, tmpdir)

        assert update_obj.update("fastq_align_bowtie2") is True
        assert cmp_component(tmpdir, sw_path) is False

        # Check that the modules.json is correctly updated
        mod_json = ModulesJson(self.pipeline_dir).get_modules_json()
        # Get the up-to-date git_sha for the sw from the ModulesRepo object
        assert (
            old_mod_json["repos"][NF_CORE_MODULES_REMOTE]["subworkflows"][NF_CORE_MODULES_NAME]["fastq_align_bowtie2"][
                "git_sha"
            ]
            != mod_json["repos"][NF_CORE_MODULES_REMOTE]["subworkflows"][NF_CORE_MODULES_NAME]["fastq_align_bowtie2"][
                "git_sha"
            ]
        )

    # Mock questionary answer: update components
    @mock.patch.object(questionary.Question, "unsafe_ask", return_value=True)
    def test_install_at_hash_and_update_limit_output(self, mock_prompt):
        """Installs an old version of a subworkflow in the pipeline and updates it with limit_output=True"""
        self.caplog.set_level(logging.INFO)
        assert self.subworkflow_install_old.install("fastq_align_bowtie2")

        update_obj = SubworkflowUpdate(self.pipeline_dir, show_diff=True, update_deps=True, limit_output=True)

        assert update_obj.update("fastq_align_bowtie2")

        # Check changes not shown for non-.nf files
        assert "Changes in 'fastq_align_bowtie2/meta.yml' but not shown" in self.caplog.text
        assert "Changes in 'bam_sort_stats_samtools/meta.yml' but not shown" in self.caplog.text
        assert "Changes in 'bam_stats_samtools/meta.yml' but not shown" in self.caplog.text
        assert "Changes in 'samtools/flagstat/meta.yml' but not shown" in self.caplog.text
        # Check changes only shown for main.nf files
        assert "Changes in 'fastq_align_bowtie2/main.nf'" in self.caplog.text
        for line in self.caplog.text.split("\n"):
            if line.startswith("---"):
                assert line.endswith("main.nf")

    def test_install_at_hash_and_update_and_save_diff_to_file(self):
        """Installs an old version of a sw in the pipeline and updates it. Save differences to a file."""
        assert self.subworkflow_install_old.install("fastq_align_bowtie2")
        patch_path = Path(self.pipeline_dir, "fastq_align_bowtie2.patch")
        update_obj = SubworkflowUpdate(self.pipeline_dir, save_diff_fn=patch_path, update_deps=True)

        # Copy the sw files and check that they are affected by the update
        tmpdir = Path(tempfile.TemporaryDirectory().name)

        sw_path = Path(self.pipeline_dir, "subworkflows", NF_CORE_MODULES_NAME, "fastq_align_bowtie2")
        shutil.copytree(sw_path, tmpdir)

        assert update_obj.update("fastq_align_bowtie2") is True
        assert cmp_component(tmpdir, sw_path) is True

        with open(patch_path) as fh:
            line = fh.readline()
            assert line.startswith(
                "Changes in component 'nf-core/fastq_align_bowtie2' between (f3c078809a2513f1c95de14f6633fe1f03572fdb) and"
            )

    def test_install_at_hash_and_update_and_save_diff_limit_output(self):
        """Installs an old version of a sw in the pipeline and updates it. Save differences to a file."""
        # Install old version of fastq_align_bowtie2
        self.subworkflow_install_old.install("fastq_align_bowtie2")
        patch_path = Path(self.pipeline_dir, "fastq_align_bowtie2.patch")
        # Update saving the differences to a patch file and with `limit_output`
        update_obj = SubworkflowUpdate(self.pipeline_dir, save_diff_fn=patch_path, update_deps=True, limit_output=True)
        assert update_obj.update("fastq_align_bowtie2")

        # Check that the patch file was created
        assert patch_path.exists(), f"Patch file was not created at {patch_path}"

        # Read the contents of the patch file
        with open(patch_path) as fh:
            content = fh.read()
            # Check changes not shown for non-.nf files
            assert "Changes in 'fastq_align_bowtie2/meta.yml' but not shown" in content
            assert "Changes in 'bam_sort_stats_samtools/meta.yml' but not shown" in content
            assert "Changes in 'bam_stats_samtools/meta.yml' but not shown" in content
            assert "Changes in 'samtools/flagstat/meta.yml' but not shown" in content
            # Check changes only shown for main.nf files
            assert "Changes in 'fastq_align_bowtie2/main.nf'" in content
            for line in content:
                if line.startswith("---"):
                    assert line.endswith("main.nf")

    def test_update_all(self):
        """Updates all subworkflows present in the pipeline"""
        # Install subworkflows fastq_align_bowtie2, bam_sort_stats_samtools, bam_stats_samtools
        self.subworkflow_install.install("fastq_align_bowtie2")
        # Update all subworkflows
        update_obj = SubworkflowUpdate(self.pipeline_dir, update_all=True, show_diff=False)
        assert update_obj.update() is True

        # We must reload the modules.json to get the updated version
        mod_json_obj = ModulesJson(self.pipeline_dir)
        mod_json = mod_json_obj.get_modules_json()
        # Loop through all subworkflows and check that they are updated (according to the modules.json file)
        for sw in mod_json["repos"][NF_CORE_MODULES_REMOTE]["subworkflows"][NF_CORE_MODULES_NAME]:
            correct_git_sha = list(update_obj.modules_repo.get_component_git_log(sw, "subworkflows", depth=1))[0][
                "git_sha"
            ]
            current_git_sha = mod_json["repos"][NF_CORE_MODULES_REMOTE]["subworkflows"][NF_CORE_MODULES_NAME][sw][
                "git_sha"
            ]
            assert correct_git_sha == current_git_sha

    def test_update_with_config_fixed_version(self):
        """Try updating when there are entries in the .nf-core.yml"""
        # Install subworkflow at the latest version
        assert self.subworkflow_install.install("fastq_align_bowtie2")

        # Fix the subworkflow version in the .nf-core.yml to an old version
        update_config = {NF_CORE_MODULES_REMOTE: {NF_CORE_MODULES_NAME: {"fastq_align_bowtie2": OLD_SUBWORKFLOWS_SHA}}}
        config_fn, tools_config = nf_core.utils.load_tools_config(self.pipeline_dir)
        setattr(tools_config, "update", update_config)
        assert config_fn is not None and tools_config is not None  # mypy
        with open(Path(self.pipeline_dir, config_fn), "w") as f:
            yaml.dump(tools_config.model_dump(), f)

        # Update all subworkflows in the pipeline
        update_obj = SubworkflowUpdate(self.pipeline_dir, update_all=True, show_diff=False)
        assert update_obj.update() is True

        # Check that the git sha for fastq_align_bowtie2 is correctly downgraded
        mod_json = ModulesJson(self.pipeline_dir).get_modules_json()
        assert "fastq_align_bowtie2" in mod_json["repos"][NF_CORE_MODULES_REMOTE]["subworkflows"][NF_CORE_MODULES_NAME]
        assert (
            "git_sha"
            in mod_json["repos"][NF_CORE_MODULES_REMOTE]["subworkflows"][NF_CORE_MODULES_NAME]["fastq_align_bowtie2"]
        )
        assert (
            mod_json["repos"][NF_CORE_MODULES_REMOTE]["subworkflows"][NF_CORE_MODULES_NAME]["fastq_align_bowtie2"][
                "git_sha"
            ]
            == OLD_SUBWORKFLOWS_SHA
        )

    def test_update_with_config_dont_update(self):
        """Try updating when sw is to be ignored"""
        # Install an old version of fastq_align_bowtie2
        self.subworkflow_install_old.install("fastq_align_bowtie2")

        # Set the fastq_align_bowtie2 field to no update in the .nf-core.yml
        update_config = {NF_CORE_MODULES_REMOTE: {NF_CORE_MODULES_NAME: {"fastq_align_bowtie2": False}}}
        config_fn, tools_config = nf_core.utils.load_tools_config(self.pipeline_dir)
        setattr(tools_config, "update", update_config)
        assert config_fn is not None and tools_config is not None  # mypy
        with open(Path(self.pipeline_dir, config_fn), "w") as f:
            yaml.dump(tools_config.model_dump(), f)

        # Update all modules in the pipeline
        update_obj = SubworkflowUpdate(self.pipeline_dir, update_all=True, show_diff=False)
        assert update_obj.update() is True

        # Check that the git sha for fastq_align_bowtie2 is correctly downgraded
        mod_json = ModulesJson(self.pipeline_dir).get_modules_json()
        assert "fastq_align_bowtie2" in mod_json["repos"][NF_CORE_MODULES_REMOTE]["subworkflows"][NF_CORE_MODULES_NAME]
        assert (
            "git_sha"
            in mod_json["repos"][NF_CORE_MODULES_REMOTE]["subworkflows"][NF_CORE_MODULES_NAME]["fastq_align_bowtie2"]
        )
        assert (
            mod_json["repos"][NF_CORE_MODULES_REMOTE]["subworkflows"][NF_CORE_MODULES_NAME]["fastq_align_bowtie2"][
                "git_sha"
            ]
            == OLD_SUBWORKFLOWS_SHA
        )

    def test_update_with_config_fix_all(self):
        """Fix the version of all nf-core subworkflows"""
        # Install subworkflow at the latest version
        assert self.subworkflow_install.install("fastq_align_bowtie2")

        # Fix the version of all nf-core subworkflows in the .nf-core.yml to an old version
        update_config = {NF_CORE_MODULES_REMOTE: OLD_SUBWORKFLOWS_SHA}
        config_fn, tools_config = nf_core.utils.load_tools_config(self.pipeline_dir)
        setattr(tools_config, "update", update_config)
        assert config_fn is not None and tools_config is not None  # mypy
        with open(Path(self.pipeline_dir, config_fn), "w") as f:
            yaml.dump(tools_config.model_dump(), f)

        # Update fastq_align_bowtie2
        update_obj = SubworkflowUpdate(self.pipeline_dir, update_all=False, update_deps=True, show_diff=False)
        assert update_obj.update("fastq_align_bowtie2") is True

        # Check that the git sha for fastq_align_bowtie2 is correctly downgraded
        mod_json = ModulesJson(self.pipeline_dir).get_modules_json()
        assert (
            "git_sha"
            in mod_json["repos"][NF_CORE_MODULES_REMOTE]["subworkflows"][NF_CORE_MODULES_NAME]["fastq_align_bowtie2"]
        )
        assert (
            mod_json["repos"][NF_CORE_MODULES_REMOTE]["subworkflows"][NF_CORE_MODULES_NAME]["fastq_align_bowtie2"][
                "git_sha"
            ]
            == OLD_SUBWORKFLOWS_SHA
        )

    def test_update_with_config_no_updates(self):
        """Don't update any nf-core subworkflows"""
        # Install an old version of fastq_align_bowtie2
        self.subworkflow_install_old.install("fastq_align_bowtie2")
        old_mod_json = ModulesJson(self.pipeline_dir).get_modules_json()

        # Set all repository updates to False
        update_config = {NF_CORE_MODULES_REMOTE: False}
        config_fn, tools_config = nf_core.utils.load_tools_config(self.pipeline_dir)
        setattr(tools_config, "update", update_config)
        assert config_fn is not None and tools_config is not None  # mypy
        with open(Path(self.pipeline_dir, config_fn), "w") as f:
            yaml.dump(tools_config.model_dump(), f)

        # Update all subworkflows in the pipeline
        update_obj = SubworkflowUpdate(self.pipeline_dir, update_all=True, show_diff=False)
        assert update_obj.update() is True

        # Check that the git sha for fastq_align_bowtie2 is correctly downgraded and none of the subworkflows has changed
        mod_json = ModulesJson(self.pipeline_dir).get_modules_json()
        for sw in mod_json["repos"][NF_CORE_MODULES_REMOTE]["subworkflows"][NF_CORE_MODULES_NAME]:
            assert "git_sha" in mod_json["repos"][NF_CORE_MODULES_REMOTE]["subworkflows"][NF_CORE_MODULES_NAME][sw]
            assert (
                mod_json["repos"][NF_CORE_MODULES_REMOTE]["subworkflows"][NF_CORE_MODULES_NAME][sw]["git_sha"]
                == old_mod_json["repos"][NF_CORE_MODULES_REMOTE]["subworkflows"][NF_CORE_MODULES_NAME][sw]["git_sha"]
            )

    def test_update_all_linked_components_from_subworkflow(self):
        """Update a subworkflow and all modules and subworkflows used on it"""
        # Install an old version of fastq_align_bowtie2
        self.subworkflow_install_old.install("fastq_align_bowtie2")
        old_mod_json = ModulesJson(self.pipeline_dir).get_modules_json()

        # Copy the sw files and check that they are affected by the update
        tmpdir = Path(tempfile.TemporaryDirectory().name)
        subworkflows_path = Path(self.pipeline_dir, "subworkflows", NF_CORE_MODULES_NAME)
        modules_path = Path(self.pipeline_dir, "modules", NF_CORE_MODULES_NAME)
        shutil.copytree(subworkflows_path, Path(tmpdir, "subworkflows"))
        shutil.copytree(modules_path, Path(tmpdir, "modules"))

        # Update fastq_align_bowtie2 and all modules and subworkflows used by that
        update_obj = SubworkflowUpdate(self.pipeline_dir, update_deps=True, show_diff=False)
        assert update_obj.update("fastq_align_bowtie2") is True

        mod_json = ModulesJson(self.pipeline_dir).get_modules_json()
        # Loop through all modules and subworkflows used in fastq_align_bowtie2
        # check that they are updated (according to the modules.json file)
        for sw in ["fastq_align_bowtie2", "bam_sort_stats_samtools", "bam_stats_samtools"]:
            assert (
                old_mod_json["repos"][NF_CORE_MODULES_REMOTE]["subworkflows"][NF_CORE_MODULES_NAME][sw]["git_sha"]
                != mod_json["repos"][NF_CORE_MODULES_REMOTE]["subworkflows"][NF_CORE_MODULES_NAME][sw]["git_sha"]
            )
        for mod in [
            "bowtie2/align",
            "samtools/index",
            "samtools/sort",
            "samtools/flagstat",
            "samtools/idxstats",
            "samtools/stats",
        ]:
            assert (
                old_mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"][NF_CORE_MODULES_NAME][mod]["git_sha"]
                != mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"][NF_CORE_MODULES_NAME][mod]["git_sha"]
            )
        # Check that the subworkflow files are updated
        assert (
            cmp_component(
                Path(tmpdir, "subworkflows", "fastq_align_bowtie2"), Path(subworkflows_path, "fastq_align_bowtie2")
            )
            is False
        )

    def test_update_all_subworkflows_from_module(self):
        """Update a module and all subworkflows that use this module"""
        # Install an old version of fastq_align_bowtie2 and thus all modules used by it (bowtie2/align)
        self.subworkflow_install_old.install("fastq_align_bowtie2")
        old_mod_json = ModulesJson(self.pipeline_dir).get_modules_json()

        # Copy the sw files and check that they are affected by the update
        tmpdir = Path(tempfile.TemporaryDirectory().name)
        sw_path = Path(self.pipeline_dir, "subworkflows", NF_CORE_MODULES_NAME, "fastq_align_bowtie2")
        shutil.copytree(sw_path, Path(tmpdir, "fastq_align_bowtie2"))

        # Update bowtie2/align and all subworkflows using it
        update_obj = ModuleUpdate(self.pipeline_dir, update_deps=True, show_diff=False)
        assert update_obj.update("bowtie2/align") is True

        mod_json = ModulesJson(self.pipeline_dir).get_modules_json()
        # Check that bowtie2/align and fastq_align_bowtie2 are updated
        assert (
            old_mod_json["repos"][NF_CORE_MODULES_REMOTE]["subworkflows"][NF_CORE_MODULES_NAME]["fastq_align_bowtie2"][
                "git_sha"
            ]
            != mod_json["repos"][NF_CORE_MODULES_REMOTE]["subworkflows"][NF_CORE_MODULES_NAME]["fastq_align_bowtie2"][
                "git_sha"
            ]
        )
        assert cmp_component(Path(tmpdir, "fastq_align_bowtie2"), sw_path) is False
        assert (
            old_mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"][NF_CORE_MODULES_NAME]["bowtie2/align"]["git_sha"]
            != mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"][NF_CORE_MODULES_NAME]["bowtie2/align"]["git_sha"]
        )

    def test_update_change_of_included_modules(self):
        """Update a subworkflow which has a module change in the new version."""
        # Install an old version of vcf_annotate_ensemblvep with tabix/bgziptabix and without tabix/tabix
        self.subworkflow_install_module_change.install("vcf_annotate_ensemblvep")
        old_mod_json = ModulesJson(self.pipeline_dir).get_modules_json()

        # Check that tabix/bgziptabix is there
        assert "tabix/bgziptabix" in old_mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"][NF_CORE_MODULES_NAME]
        assert Path(self.pipeline_dir, "modules", NF_CORE_MODULES_NAME, "tabix/bgziptabix").is_dir()
        # Check that tabix/tabix is not there
        assert "tabix/tabix" not in old_mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"][NF_CORE_MODULES_NAME]
        assert not Path(self.pipeline_dir, "modules", NF_CORE_MODULES_NAME, "tabix/tabix").is_dir()

        # Update vcf_annotate_ensemblvep without tabix/bgziptabix and with tabix/tabix
        update_obj = SubworkflowUpdate(self.pipeline_dir, update_deps=True, show_diff=False)
        assert update_obj.update("vcf_annotate_ensemblvep") is True

        mod_json = ModulesJson(self.pipeline_dir).get_modules_json()

        # Check that tabix/bgziptabix is not there
        assert "tabix/bgziptabix" not in mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"][NF_CORE_MODULES_NAME]
        assert not Path(self.pipeline_dir, "modules", NF_CORE_MODULES_NAME, "tabix/bgziptabix").is_dir()
        # Check that tabix/tabix is there
        assert "tabix/tabix" in mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"][NF_CORE_MODULES_NAME]
        assert Path(self.pipeline_dir, "modules", NF_CORE_MODULES_NAME, "tabix/tabix").is_dir()
        # Check that ensemblevep is not there but instead we have ensemblevep/vep (due to a file re-naming)
        assert "ensemblvep" not in mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"][NF_CORE_MODULES_NAME]
        assert "ensemblvep/vep" in mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"][NF_CORE_MODULES_NAME]
        assert Path(self.pipeline_dir, "modules", NF_CORE_MODULES_NAME, "ensemblvep/vep").is_dir()



import mimetypes
import re


def template_strings(self):
    """Check for template placeholders.

    The ``nf-core pipelines create`` pipeline template uses
    `Jinja <https://jinja.palletsprojects.com/en/2.11.x/>`_ behind the scenes.

    This lint test fails if any Jinja template variables such as
    ``{{ pipeline_name }}`` are found in your pipeline code.

    Finding a placeholder like this means that something was probably copied and pasted
    from the template without being properly rendered for your pipeline.

    This test ignores any double-brackets prefixed with a dollar sign, such as
    ``${{ secrets.AWS_ACCESS_KEY_ID }}`` as these placeholders are used in GitHub Actions workflows.

    .. tip:: You can choose to ignore lint test tests by editing the file called
        ``.nf-core.yml`` in the root of your pipeline and setting the test to false:

        .. code-block:: yaml

            lint:
                template_strings: False

        To disable this test only for specific files, you can specify a list of file paths to ignore.
        For example, to ignore a pdf you added to the docs:

        .. code-block:: yaml

            lint:
                template_strings:
                    - docs/my_pdf.pdf
    """
    passed = []
    failed = []
    ignored = []
    # Files that should be ignored according to the linting config
    ignore_files = self.lint_config.get("template_strings", []) if self.lint_config is not None else []

    files = self.list_files()
    # Loop through files, searching for string
    num_matches = 0
    for fn in files:
        if str(fn.relative_to(self.wf_path)) in ignore_files:
            ignored.append(f"Ignoring Jinja template strings in file `{fn}`")
            continue
        # Skip binary files
        binary_ftypes = ["image", "application/java-archive"]
        (ftype, encoding) = mimetypes.guess_type(fn)
        if encoding is not None or (ftype is not None and any([ftype.startswith(ft) for ft in binary_ftypes])):
            continue

        with open(fn, encoding="latin1") as fh:
            lnum = 0
            for line in fh:
                lnum += 1
                cc_matches = re.findall(r"[^$]{{[^:}]*}}", line)
                if len(cc_matches) > 0:
                    for cc_match in cc_matches:
                        failed.append(f"Found a Jinja template string in `{fn}` L{lnum}: {cc_match}")
                        num_matches += 1
    if num_matches == 0:
        passed.append(f"Did not find any Jinja template strings ({len(self.files)} files)")

    return {"passed": passed, "failed": failed, "ignored": ignored}



import logging
import os
import shutil
from pathlib import Path
from typing import Optional

import git
import rich
import rich.progress
import rich.prompt
from git.exc import GitCommandError, InvalidGitRepositoryError

import nf_core.modules.modules_json
import nf_core.modules.modules_utils
from nf_core.components.components_utils import NF_CORE_MODULES_NAME, NF_CORE_MODULES_REMOTE
from nf_core.synced_repo import RemoteProgressbar, SyncedRepo
from nf_core.utils import NFCORE_CACHE_DIR, NFCORE_DIR, load_tools_config

log = logging.getLogger(__name__)


class ModulesRepo(SyncedRepo):
    """
    An object to store details about the repository being used for modules.

    Used by the `nf-core modules` top-level command with -r and -b flags,
    so that this can be used in the same way by all sub-commands.

    We keep track of the pull-status of the different installed repos in
    the static variable local_repo_status. This is so we don't need to
    pull a remote several times in one command.
    """

    local_repo_statuses = {}
    no_pull_global = False

    def __init__(
        self,
        remote_url: Optional[str] = None,
        branch: Optional[str] = None,
        no_pull: bool = False,
        hide_progress: bool = False,
    ) -> None:
        """
        Initializes the object and clones the git repository if it is not already present
        """

        # This allows us to set this one time and then keep track of the user's choice
        ModulesRepo.no_pull_global |= no_pull

        # Check if the remote seems to be well formed
        if remote_url is None:
            remote_url = NF_CORE_MODULES_REMOTE

        self.remote_url = remote_url

        self.fullname = nf_core.modules.modules_utils.repo_full_name_from_remote(self.remote_url)

        self.setup_local_repo(remote_url, branch, hide_progress)

        config_fn, repo_config = load_tools_config(self.local_repo_dir)
        if config_fn is None or repo_config is None:
            raise UserWarning(f"Could not find a configuration file in {self.local_repo_dir}")
        try:
            self.repo_path = repo_config.org_path
        except KeyError:
            raise UserWarning(f"'org_path' key not present in {config_fn.name}")

        # Verify that the repo seems to be correctly configured
        if self.repo_path != NF_CORE_MODULES_NAME or self.branch:
            self.verify_branch()
        if self.repo_path is None:
            raise UserWarning(f"Could not find the org_path in the configuration file: {config_fn.name}")
        # Convenience variable
        self.modules_dir = Path(self.local_repo_dir, "modules", self.repo_path)
        self.subworkflows_dir = Path(self.local_repo_dir, "subworkflows", self.repo_path)

        self.avail_module_names = None

    def gitless_repo(self):
        gitless_repo_url = self.remote_url
        if self.remote_url and ".git" in self.remote_url:
            gitless_repo_url = gitless_repo_url[:-4]
        return gitless_repo_url

    def setup_local_repo(self, remote, branch, hide_progress=True, in_cache=False):
        """
        Sets up the local git repository. If the repository has been cloned previously, it
        returns a git.Repo object of that clone. Otherwise it tries to clone the repository from
        the provided remote URL and returns a git.Repo of the new clone.

        Args:
            remote (str): git url of remote
            branch (str): name of branch to use
        Sets self.repo
        """
        self.local_repo_dir = Path(NFCORE_DIR if not in_cache else NFCORE_CACHE_DIR, self.fullname)
        try:
            if not os.path.exists(self.local_repo_dir):
                try:
                    pbar = rich.progress.Progress(
                        "[bold blue]{task.description}",
                        rich.progress.BarColumn(bar_width=None),
                        "[bold yellow]{task.fields[state]}",
                        transient=True,
                        disable=hide_progress or os.environ.get("HIDE_PROGRESS", None) is not None,
                    )
                    with pbar:
                        self.repo = git.Repo.clone_from(
                            remote,
                            self.local_repo_dir,
                            progress=RemoteProgressbar(pbar, self.fullname, self.remote_url, "Cloning"),
                        )
                    ModulesRepo.update_local_repo_status(self.fullname, True)
                except GitCommandError:
                    raise LookupError(f"Failed to clone from the remote: `{remote}`")
                # Verify that the requested branch exists by checking it out
                self.setup_branch(branch)
            else:
                self.repo = git.Repo(self.local_repo_dir)

                if ModulesRepo.no_pull_global:
                    ModulesRepo.update_local_repo_status(self.fullname, True)
                # If the repo is already cloned, fetch the latest changes from the remote
                if not ModulesRepo.local_repo_synced(self.fullname):
                    pbar = rich.progress.Progress(
                        "[bold blue]{task.description}",
                        rich.progress.BarColumn(bar_width=None),
                        "[bold yellow]{task.fields[state]}",
                        transient=True,
                        disable=hide_progress or os.environ.get("HIDE_PROGRESS", None) is not None,
                    )
                    with pbar:
                        self.repo.remotes.origin.fetch(
                            progress=RemoteProgressbar(pbar, self.fullname, self.remote_url, "Pulling")
                        )
                    ModulesRepo.update_local_repo_status(self.fullname, True)

                # Before verifying the branch, fetch the changes
                # Verify that the requested branch exists by checking it out
                self.setup_branch(branch)

                # Now merge the changes
                tracking_branch = self.repo.active_branch.tracking_branch()
                if tracking_branch is None:
                    raise LookupError(f"There is no remote tracking branch '{self.branch}' in '{self.remote_url}'")
                self.repo.git.merge(tracking_branch.name)
        except (GitCommandError, InvalidGitRepositoryError) as e:
            log.error(f"[red]Could not set up local cache of modules repository:[/]\n{e}\n")
            if rich.prompt.Confirm.ask(f"[violet]Delete local cache '{self.local_repo_dir}' and try again?"):
                log.info(f"Removing '{self.local_repo_dir}'")
                shutil.rmtree(self.local_repo_dir)
                self.setup_local_repo(remote, branch, hide_progress)
            else:
                raise LookupError("Exiting due to error with local modules git repo")



import logging
from pathlib import Path
from typing import Any, Dict, List

import jsonschema
import requests
import yaml


def actions_schema_validation(self) -> Dict[str, List[str]]:
    """Checks that the GitHub Action workflow yml/yaml files adhere to the correct schema

    nf-core pipelines use GitHub actions workflows to run CI tests, check formatting and also linting, among others.
    These workflows are defined by ``yml`` scripts in ``.github/workflows/``. This lint test verifies that these scripts are valid
    by comparing them against the `JSON schema for GitHub workflows <https://json.schemastore.org/github-workflow>`_.

    To pass this test, make sure that all your workflows contain the required properties ``on`` and ``jobs`` and that
    all other properties are of the correct type, as specified in the schema (link above).
    """
    passed: List[str] = []
    failed: List[str] = []
    warned: List[str] = []

    # Only show error messages from schema
    logging.getLogger("nf_core.pipelines.schema").setLevel(logging.ERROR)

    # Get all workflow files
    action_workflows = list(Path(self.wf_path).glob(".github/workflows/*.y*ml"))

    # Load the GitHub workflow schema
    r = requests.get("https://json.schemastore.org/github-workflow", allow_redirects=True)
    # handle "Service Unavailable" error
    if r.status_code not in [200, 301]:
        warned.append(
            f"Failed to fetch schema: Response code for `https://json.schemastore.org/github-workflow` was {r.status_code}"
        )
        return {"passed": passed, "failed": failed, "warned": warned}
    schema: Dict[str, Any] = r.json()

    # Validate all workflows against the schema
    for wf_path in action_workflows:
        wf = wf_path.name

        # load workflow
        try:
            with open(wf_path) as fh:
                wf_json = yaml.safe_load(fh)
        except Exception as e:
            failed.append(f"Could not parse yaml file: {wf}, {e}")
            continue

        # yaml parses 'on' as True --> try to fix it before schema validation
        try:
            wf_json["on"] = wf_json.pop(True)
        except Exception:
            failed.append(f"Missing 'on' keyword in {wf}")

        # Validate the workflow
        try:
            jsonschema.validate(wf_json, schema)
            passed.append(f"Workflow validation passed: {wf}")
        except Exception as e:
            failed.append(f"Workflow validation failed for {wf}: {e}")

    return {"passed": passed, "failed": failed, "warned": warned}



from pathlib import Path
from typing import Dict, List

import yaml

from nf_core.pipelines.lint_utils import ignore_file


def multiqc_config(self) -> Dict[str, List[str]]:
    """Make sure basic multiQC plugins are installed and plots are exported
    Basic template:

    .. code-block:: yaml

        report_comment: >
            This report has been generated by the <a href="https://github.com/nf-core/quantms" target="_blank">nf-core/quantms</a>
            analysis pipeline. For information about how to interpret these results, please see the
            <a href="https://nf-co.re/quantms" target="_blank">documentation</a>.
        report_section_order:
            software_versions:
                order: -1000
            nf-core-quantms-summary:
                order: -1001
        export_plots: true

    .. note:: You can choose to ignore this lint tests by editing the file called
        ``.nf-core.yml`` in the root of your pipeline and setting the test to false:

        .. code-block:: yaml

            lint:
                multiqc_config: False

        To disable this test only for specific sections, you can specify a list of section names.
        For example:

        .. code-block:: yaml
        lint:
                multiqc_config:
                    - report_section_order
                    - report_comment

    """

    passed: List[str] = []
    failed: List[str] = []
    ignored: List[str] = []

    fn = Path(self.wf_path, "assets", "multiqc_config.yml")
    file_path = fn.relative_to(self.wf_path)
    passed, failed, ignored, ignore_configs = ignore_file("multiqc_config", file_path, self.wf_path)

    # skip other tests if the file is not found
    error_message = f"`{file_path}` not found"
    # check for partial match in failed or ignored
    if not any(f.startswith(error_message) for f in (failed + ignored)):
        try:
            with open(fn) as fh:
                mqc_yml = yaml.safe_load(fh)
        except Exception as e:
            return {"failed": [f"Could not parse yaml file: {fn}, {e}"]}

        # check if required sections are present
        required_sections = ["report_section_order", "export_plots", "report_comment"]
        for section in required_sections:
            if section not in mqc_yml and section not in ignore_configs:
                failed.append(f"`assets/multiqc_config.yml` does not contain `{section}`")
                return {"passed": passed, "failed": failed}
            else:
                passed.append(f"`assets/multiqc_config.yml` contains `{section}`")

        try:
            orders = {}
            summary_plugin_name = f"{self.pipeline_prefix}-{self.pipeline_name}-summary"
            min_plugins = ["software_versions", summary_plugin_name]
            for plugin in min_plugins:
                if plugin not in mqc_yml["report_section_order"]:
                    raise AssertionError(f"Section {plugin} missing in report_section_order")
                if "order" not in mqc_yml["report_section_order"][plugin]:
                    raise AssertionError(f"Section {plugin} 'order' missing. Must be < 0")
                plugin_order = mqc_yml["report_section_order"][plugin]["order"]
                if plugin_order >= 0:
                    raise AssertionError(f"Section {plugin} 'order' must be < 0")

            for plugin in mqc_yml["report_section_order"]:
                if "order" in mqc_yml["report_section_order"][plugin]:
                    orders[plugin] = mqc_yml["report_section_order"][plugin]["order"]

            if orders[summary_plugin_name] != min(orders.values()):
                raise AssertionError(f"Section {summary_plugin_name} should have the lowest order")
            orders.pop(summary_plugin_name)
            if orders["software_versions"] != min(orders.values()):
                raise AssertionError("Section software_versions should have the second lowest order")
        except (AssertionError, KeyError, TypeError) as e:
            failed.append(f"`assets/multiqc_config.yml` does not meet requirements: {e}")
        else:
            passed.append("`assets/multiqc_config.yml` follows the ordering scheme of the minimally required plugins.")

        if "report_comment" not in ignore_configs:
            # Check that the minimum plugins exist and are coming first in the summary
            version = self.nf_config.get("manifest.version", "").strip(" '\"")
            if "dev" in version:
                version = "dev"
                report_comments = (
                    f'This report has been generated by the <a href="https://github.com/nf-core/{self.pipeline_name}/tree/dev" target="_blank">nf-core/{self.pipeline_name}</a>'
                    f" analysis pipeline. For information about how to interpret these results, please see the "
                    f'<a href="https://nf-co.re/{self.pipeline_name}/dev/docs/output" target="_blank">documentation</a>.'
                )

            else:
                report_comments = (
                    f'This report has been generated by the <a href="https://github.com/nf-core/{self.pipeline_name}/releases/tag/{version}" target="_blank">nf-core/{self.pipeline_name}</a>'
                    f" analysis pipeline. For information about how to interpret these results, please see the "
                    f'<a href="https://nf-co.re/{self.pipeline_name}/{version}/docs/output" target="_blank">documentation</a>.'
                )

            if mqc_yml["report_comment"].strip() != report_comments:
                # find where the report_comment is wrong and give it as a hint
                hint = report_comments
                failed.append(
                    f"`assets/multiqc_config.yml` does not contain a matching 'report_comment'.  \n"
                    f"The expected comment is:  \n"
                    f"```{hint}```  \n"
                    f"The current comment is:  \n"
                    f"```{ mqc_yml['report_comment'].strip()}```"
                )
            else:
                passed.append("`assets/multiqc_config.yml` contains a matching 'report_comment'.")

        # Check that export_plots is activated
        try:
            if not mqc_yml["export_plots"]:
                raise AssertionError()
        except (AssertionError, KeyError, TypeError):
            failed.append("`assets/multiqc_config.yml` does not contain 'export_plots: true'.")
        else:
            passed.append("`assets/multiqc_config.yml` contains 'export_plots: true'.")

    return {"passed": passed, "failed": failed, "ignored": ignored}



import os
import shutil
from pathlib import Path
from unittest import mock

import pytest
import requests_cache
import responses
import yaml
from git.repo import Repo

import nf_core.modules.create
from tests.utils import (
    GITLAB_SUBWORKFLOWS_ORG_PATH_BRANCH,
    GITLAB_URL,
    mock_anaconda_api_calls,
    mock_biocontainers_api_calls,
)

from ..test_modules import TestModules


class TestModulesCreate(TestModules):
    def test_modules_create_succeed(self):
        """Succeed at creating the TrimGalore! module"""
        with responses.RequestsMock() as rsps:
            mock_anaconda_api_calls(rsps, "trim-galore", "0.6.7")
            mock_biocontainers_api_calls(rsps, "trim-galore", "0.6.7")
            module_create = nf_core.modules.create.ModuleCreate(
                self.pipeline_dir, "trimgalore", "@author", "process_single", True, True, conda_name="trim-galore"
            )
            with requests_cache.disabled():
                module_create.create()
        assert os.path.exists(os.path.join(self.pipeline_dir, "modules", "local", "trimgalore.nf"))

    def test_modules_create_fail_exists(self):
        """Fail at creating the same module twice"""
        with responses.RequestsMock() as rsps:
            mock_anaconda_api_calls(rsps, "trim-galore", "0.6.7")
            mock_biocontainers_api_calls(rsps, "trim-galore", "0.6.7")
            module_create = nf_core.modules.create.ModuleCreate(
                self.pipeline_dir, "trimgalore", "@author", "process_single", False, False, conda_name="trim-galore"
            )
            with requests_cache.disabled():
                module_create.create()
            with pytest.raises(UserWarning) as excinfo:
                with requests_cache.disabled():
                    module_create.create()
        assert "Module file exists already" in str(excinfo.value)

    def test_modules_create_nfcore_modules(self):
        """Create a module in nf-core/modules clone"""
        with responses.RequestsMock() as rsps:
            mock_anaconda_api_calls(rsps, "fastqc", "0.11.9")
            mock_biocontainers_api_calls(rsps, "fastqc", "0.11.9")
            module_create = nf_core.modules.create.ModuleCreate(
                self.nfcore_modules, "fastqc", "@author", "process_low", False, False
            )
            with requests_cache.disabled():
                module_create.create()
        assert os.path.exists(os.path.join(self.nfcore_modules, "modules", "nf-core", "fastqc", "main.nf"))
        assert os.path.exists(
            os.path.join(self.nfcore_modules, "modules", "nf-core", "fastqc", "tests", "main.nf.test")
        )

    def test_modules_create_nfcore_modules_subtool(self):
        """Create a tool/subtool module in a nf-core/modules clone"""
        with responses.RequestsMock() as rsps:
            mock_anaconda_api_calls(rsps, "star", "2.8.10a")
            mock_biocontainers_api_calls(rsps, "star", "2.8.10a")
            module_create = nf_core.modules.create.ModuleCreate(
                self.nfcore_modules, "star/index", "@author", "process_medium", False, False
            )
            with requests_cache.disabled():
                module_create.create()
        assert os.path.exists(os.path.join(self.nfcore_modules, "modules", "nf-core", "star", "index", "main.nf"))
        assert os.path.exists(
            os.path.join(self.nfcore_modules, "modules", "nf-core", "star", "index", "tests", "main.nf.test")
        )

    @mock.patch("rich.prompt.Confirm.ask")
    def test_modules_migrate(self, mock_rich_ask):
        """Create a module with the --migrate-pytest option to convert pytest to nf-test"""
        pytest_dir = Path(self.nfcore_modules, "tests", "modules", "nf-core", "samtools", "sort")
        module_dir = Path(self.nfcore_modules, "modules", "nf-core", "samtools", "sort")

        # Clone modules repo with pytests
        shutil.rmtree(self.nfcore_modules)
        Repo.clone_from(GITLAB_URL, self.nfcore_modules, branch=GITLAB_SUBWORKFLOWS_ORG_PATH_BRANCH)
        with open(module_dir / "main.nf") as fh:
            old_main_nf = fh.read()
        with open(module_dir / "meta.yml") as fh:
            old_meta_yml = fh.read()

        # Create a module with --migrate-pytest
        mock_rich_ask.return_value = True
        module_create = nf_core.modules.create.ModuleCreate(self.nfcore_modules, "samtools/sort", migrate_pytest=True)
        module_create.create()

        with open(module_dir / "main.nf") as fh:
            new_main_nf = fh.read()
        with open(module_dir / "meta.yml") as fh:
            new_meta_yml = fh.read()
        nextflow_config = module_dir / "tests" / "nextflow.config"

        # Check that old files have been copied to the new module
        assert old_main_nf == new_main_nf
        assert old_meta_yml == new_meta_yml
        assert nextflow_config.is_file()

        # Check that pytest folder is deleted
        assert not pytest_dir.is_dir()

        # Check that pytest_modules.yml is updated
        with open(Path(self.nfcore_modules, "tests", "config", "pytest_modules.yml")) as fh:
            modules_yml = yaml.safe_load(fh)
        assert "samtools/sort" not in modules_yml.keys()

    @mock.patch("rich.prompt.Confirm.ask")
    def test_modules_migrate_no_delete(self, mock_rich_ask):
        """Create a module with the --migrate-pytest option to convert pytest to nf-test.
        Test that pytest directory is not deleted."""
        pytest_dir = Path(self.nfcore_modules, "tests", "modules", "nf-core", "samtools", "sort")

        # Clone modules repo with pytests
        shutil.rmtree(self.nfcore_modules)
        Repo.clone_from(GITLAB_URL, self.nfcore_modules, branch=GITLAB_SUBWORKFLOWS_ORG_PATH_BRANCH)

        # Create a module with --migrate-pytest
        mock_rich_ask.return_value = False
        module_create = nf_core.modules.create.ModuleCreate(self.nfcore_modules, "samtools/sort", migrate_pytest=True)
        module_create.create()

        # Check that pytest folder is not deleted
        assert pytest_dir.is_dir()

        # Check that pytest_modules.yml is updated
        with open(Path(self.nfcore_modules, "tests", "config", "pytest_modules.yml")) as fh:
            modules_yml = yaml.safe_load(fh)
        assert "samtools/sort" not in modules_yml.keys()

    @mock.patch("rich.prompt.Confirm.ask")
    def test_modules_migrate_symlink(self, mock_rich_ask):
        """Create a module with the --migrate-pytest option to convert pytest with symlinks to nf-test.
        Test that the symlink is deleted and the file is copied."""

        pytest_dir = Path(self.nfcore_modules, "tests", "modules", "nf-core", "samtools", "sort")
        module_dir = Path(self.nfcore_modules, "modules", "nf-core", "samtools", "sort")

        # Clone modules repo with pytests
        shutil.rmtree(self.nfcore_modules)
        Repo.clone_from(GITLAB_URL, self.nfcore_modules, branch=GITLAB_SUBWORKFLOWS_ORG_PATH_BRANCH)

        # Create a symlinked file in the pytest directory
        symlink_file = pytest_dir / "symlink_file.txt"
        symlink_file.symlink_to(module_dir / "main.nf")

        # Create a module with --migrate-pytest
        mock_rich_ask.return_value = True
        module_create = nf_core.modules.create.ModuleCreate(self.nfcore_modules, "samtools/sort", migrate_pytest=True)
        module_create.create()

        # Check that symlink is deleted
        assert not symlink_file.is_symlink()
